import pytest

import mlflow_tracking_server.gunicorn.run as run


@pytest.mark.parametrize(
    "argv",
    [
        [],
        ["--testing"],
        ["--artifacts-destination", "test-destination"],
        ["--artifacts-destination", "test-destination", "--serve-artifacts"],
    ],
)
def test_run(tmp_path, argv):
    run._create_app([*argv, "--backend-store-uri", tmp_path.as_posix()])



_PROJECT_ID_ENV_VAR = "MANTIK_PROJECT_ID"
_DATA_REPOSITORY_ID_ENV_VAR = "MANTIK_DATA_REPOSITORY_ID"
_EXPERIMENT_REPOSITORY_ID_ENV_VAR = "MANTIK_EXPERIMENT_REPOSITORY_ID"
_CODE_REPOSITORY_ID_ENV_VAR = "MANTIK_CODE_REPOSITORY_ID"



import click

import mantik.cli._callbacks as _callbacks


VERBOSE = click.option(
    "--verbose",
    "-v",
    is_flag=True,
    callback=_callbacks.set_verbose_logging,
    help="Set logging to verbose mode.",
)



import logging
import pathlib
import platform
import typing as t
import uuid

import click

import mantik
import mantik.cli._options as _main_options
import mantik.cli.runs.runs as runs
import mantik.cli.utils as cli_utils
import mantik.runs.local as local_runs
import mantik.runs.schemas as schemas
import mantik.utils.env_vars as env_vars


logger = logging.getLogger(__name__)


def get_gpu_info() -> list:
    import GPUtil

    gpus = GPUtil.getGPUs()

    gpu_info_list = []

    for gpu in gpus:
        gpu_info = {
            "Name": gpu.name,
            "ID": str(gpu.id),
            "Driver": gpu.driver,
            "TotalMemory": str(gpu.memoryTotal),
        }

        gpu_info_list.append(gpu_info)

    return gpu_info_list


def get_system_details() -> dict:
    import cpuinfo
    import psutil

    system_details = {
        "Platform": platform.platform(),
        "CpuModel": cpuinfo.get_cpu_info().get("brand_raw"),
        "CpuCoreCount": psutil.cpu_count(),
        "TotalMemoryGB": psutil.virtual_memory().total / (1024**3),
        "GPUInfo": get_gpu_info(),
    }
    return {"SystemDetails": system_details}


@runs.cli.command("local")
@click.argument(
    "mlproject-path",
    type=click.Path(path_type=pathlib.Path),
    required=True,
)
@click.option(
    "--run-name",
    default=None,
    type=str,
    help="Name of the Run.",
    required=True,
)
@click.option(
    "--entry-point",
    required=False,
    default="main",
    show_default=True,
    help="Entrypoint of the MLproject file.",
)
@click.option(
    "--project-id",
    type=uuid.UUID,
    required=True,
    help=f"""Project ID on Mantik.

        If not specified, it is inferred from the environment variable
        {env_vars._PROJECT_ID_ENV_VAR}.

    """,
    envvar=env_vars._PROJECT_ID_ENV_VAR,
)
@click.option(
    "--data-repository-id",
    type=uuid.UUID,
    default=None,
    help=f"""Data repository ID on Mantik

        If not specified, it is inferred from the environment variable
        {env_vars._DATA_REPOSITORY_ID_ENV_VAR}.

    """,
    envvar=env_vars._DATA_REPOSITORY_ID_ENV_VAR,
)
@click.option(
    "--experiment-repository-id",
    type=uuid.UUID,
    required=True,
    help=f"""Experiment repository ID on Mantik

        If not specified, it is inferred from the environment variable
        {env_vars._EXPERIMENT_REPOSITORY_ID_ENV_VAR}.

    """,
    envvar=env_vars._EXPERIMENT_REPOSITORY_ID_ENV_VAR,
)
@click.option(
    "--code-repository-id",
    type=uuid.UUID,
    required=True,
    help=f"""Code repository ID on Mantik

        If not specified, it is inferred from the environment variable
        {env_vars._CODE_REPOSITORY_ID_ENV_VAR}.

    """,
    envvar=env_vars._CODE_REPOSITORY_ID_ENV_VAR,
)
@click.option(
    "--branch",
    default=None,
    type=str,
    help="Name of the code repository's branch you want to use for this run",
    required=False,
)
@click.option(
    "--commit",
    default=None,
    type=str,
    help="""Name of the code repository's full commit hash you want to use for
        this run.

        If both branch and commit hash are given, the commit hash is preferred
        over the branch.
    """,
    required=False,
)
@click.option(
    "--backend-config-system-info",
    type=bool,
    default=True,
    help="Populate the backend config JSON with system info.",
)
@click.option(
    "--parameter", "-P", show_default=True, default=lambda: [], multiple=True
)
@_main_options.VERBOSE
def run_project(
    run_name: str,
    mlproject_path: pathlib.Path,
    entry_point: str,
    parameter: t.List[str],
    verbose: bool,  # noqa
    project_id: uuid.UUID,
    data_repository_id: t.Optional[uuid.UUID],
    experiment_repository_id: uuid.UUID,
    code_repository_id: uuid.UUID,
    branch: t.Optional[str],
    commit: t.Optional[str],
    backend_config_system_info: bool,
) -> None:
    """Run an MLflow project locally and save the results in Mantik API

    Note that `MLPROJECT_PATH` is the relative path to the MLflow project file
    with your Code Repository as root.

    Remember that when you execute a run, the code is retrieved from your
    remote Git Code Repository. So make sure to commit and push your
    changes before executing a run!

    To find the respective required IDs make sure to check Mantik's UI

    """

    if branch is None and commit is None:
        raise ValueError("Either provide a branch or full commit hash to run")

    logger.debug("Parsing MLflow entry point parameters")
    parameters = cli_utils.dict_from_list(parameter)

    mantik_token = mantik.authentication.auth.get_valid_access_token()

    system_details = get_system_details() if backend_config_system_info else {}

    local_runs.run(
        data=schemas.RunConfiguration(
            name=run_name,
            experiment_repository_id=experiment_repository_id,
            code_repository_id=code_repository_id,
            branch=branch,
            commit=commit,
            data_repository_id=data_repository_id,
            mlflow_mlproject_file_path=mlproject_path.as_posix(),
            entry_point=entry_point,
            mlflow_parameters=parameters,
            backend_config=system_details,
        ),
        project_id=project_id,
        mantik_token=mantik_token,
    )
    click.echo("Done!")



import pathlib
import typing as t

import mantik.compute_backend.config.core as core
import mantik.compute_backend.config.environment as environment
import mantik.compute_backend.config.executable as executable
import mantik.compute_backend.config.resources as resources
import mantik.unicore.credentials as unicore_credentials

MLFLOW_ENV_TEST_ENV_VAR = "MLFLOW_TESTING_MANTIK_TEST_VAR"

ALL_ENV_VARS = [
    unicore_credentials._USERNAME_ENV_VAR,
    unicore_credentials._PASSWORD_ENV_VAR,
    core._PROJECT_ENV_VAR,
    MLFLOW_ENV_TEST_ENV_VAR,
]

DEFAULT_ENV_VAR_VALUE = "test-val"

EXISTING_FILE = (
    pathlib.Path(__file__).parent / "../../../resources/test-project/recipe.def"
).as_posix()


def _create_apptainer_environment(
    path: pathlib.Path,
    type: str = None,
    variables: t.Optional[t.Dict] = None,
    modules: t.Optional[t.List] = None,
    pre_run_command_on_login_node: t.Optional[str] = None,
    pre_run_command_on_compute_node: t.Optional[str] = None,
    post_run_command_on_compute_node: t.Optional[str] = None,
    post_run_command_on_login_node: t.Optional[str] = None,
    include_mlflow_env_vars: bool = True,
) -> environment.Environment:
    if type is None:
        type = "local"
    if include_mlflow_env_vars:
        variables = _include_mlflow_env_vars(variables)

    return environment.Environment(
        execution=executable.Apptainer(
            path=path,
            type=type,
        ),
        variables=variables,
        modules=modules,
        pre_run_command_on_login_node=pre_run_command_on_login_node,
        pre_run_command_on_compute_node=pre_run_command_on_compute_node,
        post_run_command_on_compute_node=post_run_command_on_compute_node,
        post_run_command_on_login_node=post_run_command_on_login_node,
    )


def _create_python_environment(
    path: pathlib.Path,
    variables: t.Optional[t.Dict] = None,
    modules: t.Optional[t.List] = None,
    pre_run_command_on_login_node: t.Optional[str] = None,
    pre_run_command_on_compute_node: t.Optional[str] = None,
    post_run_command_on_compute_node: t.Optional[str] = None,
    post_run_command_on_login_node: t.Optional[str] = None,
    include_mlflow_env_vars: bool = True,
) -> environment.Environment:
    if include_mlflow_env_vars:
        variables = _include_mlflow_env_vars(variables)

    return environment.Environment(
        executable.Python(
            path=path,
        ),
        variables=variables,
        modules=modules,
        pre_run_command_on_login_node=pre_run_command_on_login_node,
        pre_run_command_on_compute_node=pre_run_command_on_compute_node,
        post_run_command_on_compute_node=post_run_command_on_compute_node,
        post_run_command_on_login_node=post_run_command_on_login_node,
    )


def _include_mlflow_env_vars(env_variables: t.Optional[t.Dict]) -> t.Dict:
    mlflow_env_vars = {
        MLFLOW_ENV_TEST_ENV_VAR: DEFAULT_ENV_VAR_VALUE,
    }
    if env_variables is None:
        env_variables = mlflow_env_vars
    else:
        env_variables = {**env_variables, **mlflow_env_vars}
    return env_variables


def _create_config(
    resources: resources.Resources,
    api_url: str = "test-url",
    user: str = "test-val",
    password: str = "test-val",
    project: str = "test-val",
    env: t.Optional[environment.Environment] = None,
):
    if env is None:
        env = environment.Environment()
    env.variables = _include_mlflow_env_vars(env.variables)

    return core.Config(
        api_url=api_url,
        user=user,
        password=password,
        project=project,
        resources=resources,
        environment=env,
    )



import datetime
import json
import pathlib

import pytest

import mantik.compute_backend.config as config
import mantik_compute_backend.unicore._properties as _properties

FILE_DIR = pathlib.Path(__file__).parent


@pytest.fixture(scope="function")
def example_config() -> config.core.Config:
    return config.core.Config(
        api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Apptainer(
                path=pathlib.Path("mantik-test.sif"),
            ),
            variables={"SRUN_CPUS_PER_TASK": 100},
        ),
        resources=config.resources.Resources(queue="batch"),
        exclude=["*.py", "*.sif"],
    )


@pytest.fixture()
def example_job_property_response() -> dict:
    with open(
        FILE_DIR
        / "../../../resources/unicore-responses/job-property-response.json",
    ) as f:
        return json.load(f)


@pytest.fixture()
def example_job_properties():
    return _properties.Properties(
        status=_properties.Status.SUCCESSFUL,
        logs=[],
        owner="owner",
        site_name="siteName",
        consumed_time=_properties.ConsumedTime(
            total=datetime.timedelta(seconds=1),
            queued=datetime.timedelta(seconds=2),
            stage_in=datetime.timedelta(seconds=3),
            pre_command=datetime.timedelta(seconds=4),
            main=datetime.timedelta(seconds=5),
            post_command=datetime.timedelta(seconds=6),
            stage_out=datetime.timedelta(seconds=7),
        ),
        current_time=datetime.datetime(
            2000, 1, 1, tzinfo=_properties.TZ_OFFSET
        ),
        submission_time=datetime.datetime(
            2000, 1, 1, tzinfo=_properties.TZ_OFFSET
        ),
        termination_time=datetime.datetime(
            2000, 1, 2, tzinfo=_properties.TZ_OFFSET
        ),
        status_message="statusMessage",
        tags=["tag"],
        resource_status="resourceStatus",
        name="name",
        exit_code="0",
        queue="queue",
        submission_preferences={"any": "preferences"},
        resource_status_message="resourceStatusMessage",
        acl=["acl"],
        batch_system_id="batchSystemID",
    )


@pytest.fixture()
def example_config_for_python() -> config.core.Config:
    return config.core.Config(
        api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=config.environment.Environment(
            execution=config.executable.Python(
                path=pathlib.Path("/venv"),
            ),
            pre_run_command_on_compute_node="precommand compute node",
            post_run_command_on_compute_node="postcommand compute node",
        ),
        resources=config.resources.Resources(queue="batch"),
        exclude=["*.sif"],
    )



import dataclasses
import typing as t
import uuid

import jose.jwt

import mantik.authentication as authentication
import mantik.utils as utils
import mantik.utils.mantik_api.connection

_USERNAME_ENV_VAR = "MANTIK_UNICORE_USERNAME"
_PASSWORD_ENV_VAR = "MANTIK_UNICORE_PASSWORD"


@dataclasses.dataclass
class UnicoreCredentials:
    username: str
    password: str

    @classmethod
    def get_credentials(
        cls,
        connection_id: t.Optional[uuid.UUID] = None,
    ) -> "UnicoreCredentials":
        if connection_id:
            return cls._credentials_from_api(connection_id=connection_id)
        return cls._credentials_from_env_vars()

    @classmethod
    def _credentials_from_api(
        cls,
        connection_id: uuid.UUID,
    ) -> "UnicoreCredentials":
        access_token = authentication.auth.get_valid_access_token()
        user_id = _get_sub_from_token(access_token)
        connection = mantik.utils.mantik_api.connection.get(
            user_id=uuid.UUID(user_id),
            connection_id=connection_id,
            token=access_token,
        )
        return cls(username=connection.login_name, password=connection.password)

    @classmethod
    def _credentials_from_env_vars(cls) -> "UnicoreCredentials":
        username = utils.env.get_required_env_var(_USERNAME_ENV_VAR)
        password = utils.env.get_required_env_var(_PASSWORD_ENV_VAR)
        return cls(username=username, password=password)


def _get_sub_from_token(token: str):
    # The `sub` field of the token claims contains the user UUID.
    return jose.jwt.get_unverified_claims(token)["sub"]



import pytest


@pytest.mark.parametrize(
    ("max_results", "added_experiments", "allowed_to_see", "expected"),
    [
        (
            "1",
            ["Default", "test_experiment_1", "test_experiment_2"],
            ["Default"],
            ["Default"],
        ),
        (
            "3",
            ["Default", "test_experiment_1", "test_experiment_2"],
            ["Default", "test_experiment_1", "test_experiment_2"],
            ["Default", "test_experiment_1", "test_experiment_2"],
        ),
        (
            "2",
            ["Default", "test_experiment_1", "test_experiment_2"],
            ["Default", "test_experiment_1", "test_experiment_2"],
            ["test_experiment_2", "test_experiment_1"],
        ),
    ],
)
def test_get_experiment(
    client,
    mock_api_permissions,
    add_experiments_to_mlflow,
    max_results,
    added_experiments,
    allowed_to_see,
    expected,
    get_experiment_ids,
):
    add_experiments_to_mlflow(added_experiments)
    experiment_ids = get_experiment_ids(allowed_to_see)
    with mock_api_permissions(
        [
            (
                "POST",
                "mantik-api/mlflow/permissions/experiments",
                {
                    "allowed": True,
                    "message": "Allow for creating experiments and to search",
                    "experimentIds": experiment_ids,
                },
            )
        ]
    ):
        response = client.post(
            "api/2.0/mlflow/experiments/search",
            json={
                "max_results": max_results,
            },
            headers={"Authorization": "test-valid-token"},
        )
        assert response.status_code == 200
        assert sorted(
            [experiment["name"] for experiment in response.json["experiments"]]
        ) == sorted(expected)
        assert response.status_code == 200



import logging
import uuid

import mantik.utils.mantik_api.client as client

logger = logging.getLogger(__name__)


def submit_run(project_id: uuid.UUID, submit_run_data: dict, token: str):
    endpoint = f"/mantik-api/projects/{project_id}/runs"
    response = client.send_request_to_mantik_api(
        method="POST", data=submit_run_data, url_endpoint=endpoint, token=token
    )
    logger.info("Run has been successfully submitted")
    return response


def save_run(project_id: uuid.UUID, run_data: dict, token: str):
    endpoint = f"/mantik-api/projects/{project_id}/runs"
    response = client.send_request_to_mantik_api(
        method="POST",
        data=run_data,
        url_endpoint=endpoint,
        token=token,
        query_params={"submit": False},
    )
    logger.info("Run has been successfully saved")
    return response


def update_run_status(
    project_id: uuid.UUID, run_id: uuid.UUID, status: str, token: str
):
    endpoint = f"/mantik-api/projects/{project_id}/runs/{run_id}/status"
    response = client.send_request_to_mantik_api(
        method="PUT", data=status, url_endpoint=endpoint, token=token
    )
    logger.info("Run status has been successfully updated")
    return response



import uuid

import click

CONNECTION_ID = click.option(
    "--connection-id",
    required=False,
    default=None,
    type=uuid.UUID,
    help="Connection ID of your Connection stored on Mantik.",
)

DATA_REPOSITORY_NAME = click.option(
    "--name",
    required=False,
    default=None,
    type=str,
    help="Data repository name, displayed in Mantik. "
    "When not passed the TARGET is passed as name.",
)

DATA_REPOSITORY_ID = click.option(
    "--id",
    required=False,
    default=None,
    type=uuid.UUID,
    help="Data repository ID.",
)

DATA_REPOSITORY_DESCRIPTION = click.option(
    "--description",
    required=False,
    default=None,
    type=str,
    help="Data repository description, displayed in Mantik.",
)



import mantik_compute_backend.unicore._properties as _properties


class TestProperties:
    def test_from_dict(self, example_job_property_response):
        properties = _properties.Properties.from_dict(
            example_job_property_response
        )
        assert properties.status == _properties.Status.SUCCESSFUL
        assert properties.logs == [
            "Tue Oct 05 11:39:44 CEST 2021: "
            "Created with ID f0e40654-c5f9-4684-b8b1-c65ef4000125",
            "Tue Oct 05 11:39:44 CEST 2021: Created with type 'JSON'",
            "Tue Oct 05 11:39:44 CEST 2021: Client: Name: "
            "UID=test-user@emai.domain\nXlogin: uid: [test-user], "
            "gids: [test-project, addingOSgroups: true]\nRole: user: "
            "role from attribute source\nQueues: ["
            "gpus:batch:develgpus:devel:"
            "develbooster, selected=gpus]\nSecurity tokens: User: "
            "UID=test-user@email.domain\nClient's original IP: 90.189.81.115",
            "Tue Oct 05 11:39:46 CEST 2021: No staging in needed.",
            "Tue Oct 05 11:39:46 CEST 2021: Status set to READY.",
            "Tue Oct 05 11:39:46 CEST 2021: Status set to PENDING.",
            "Tue Oct 05 11:39:46 CEST 2021: No application to execute, "
            "changing action status to POSTPROCESSING",
            "Tue Oct 05 11:39:46 CEST 2021: Status set to DONE.",
            "Tue Oct 05 11:39:46 CEST 2021: Result: Success.",
            "Tue Oct 05 11:39:46 CEST 2021: Total: 0 sec., "
            "Stage-in: 0 sec., Stage-out: 0 sec.",
        ]



import dataclasses
import logging
import pathlib
import typing as t
import uuid

import mantik.compute_backend.config._base as _base
import mantik.compute_backend.config._utils as _utils
import mantik.compute_backend.config.environment as _environment
import mantik.compute_backend.config.read as read
import mantik.compute_backend.config.resources as _resources
import mantik.unicore.credentials as _unicore_credentials
import mantik.utils.env as env

_PROJECT_ENV_VAR = "MANTIK_COMPUTE_BUDGET_ACCOUNT"
_UNICORE_AUTH_SERVER_URL_ENV_VAR = "MANTIK_UNICORE_AUTH_SERVER_URL"
APPLICATION_LOGS_FILE = "mantik.log"

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Config(_base.ConfigObject):
    """The backend config for the UNICORE MLflow backend."""

    api_url: str
    user: str
    password: str
    project: str
    resources: _resources.Resources
    environment: t.Optional[_environment.Environment] = None
    exclude: t.Optional[t.List] = None

    @classmethod
    def _from_dict(
        cls, config: t.Dict, connection_id: t.Optional[uuid.UUID] = None
    ) -> "Config":
        """

        Parameters
        ----------
        config : t.Dict
            the possible keys to specify inside the config dictionary are:
            REQUIRED:
                - UnicoreApiUrl : str
                    URL to the API of the UNICORE HPC used.

                The other required field MANTIK_USERNAME, MANTIK_PASSWORD
                and MANTIK_PROJECT are inferred from the environment.
            OPTIONAL:
                - Resources : dict
                    Dict of parameters specifying the resources
                    to request on the remote system.
                    More info can be found here:
                    https://sourceforge.net/p/unicore/wiki/Job_Description/
                - Environment : dict
                    Used to build a environment.Environment.
                - Exclude : list
                    List of files or file-patterns
                    that are not sent with the job


        Returns
        -------
        mantik.unicore._config.core.Config

        """
        api_url = _utils.get_required_config_value(
            name="UnicoreApiUrl",
            value_type=str,
            config=config,
        )
        unicore_credentials = (
            _unicore_credentials.UnicoreCredentials.get_credentials(
                connection_id
            )
        )
        project = env.get_required_env_var(_PROJECT_ENV_VAR)
        resources = _resources.Resources.from_dict(
            _utils.get_required_config_value(
                name="Resources",
                value_type=dict,
                config=config,
            )
        )
        environment = _utils.get_optional_config_value(
            name="Environment",
            value_type=_environment.Environment.from_dict,
            config=config,
        )
        exclude = _utils.get_optional_config_value(
            name="Exclude",
            value_type=list,
            config=config,
        )

        return cls(
            api_url=api_url,
            user=unicore_credentials.username,
            password=unicore_credentials.password,
            project=project,
            resources=resources,
            environment=environment,
            exclude=exclude,
        )

    def __post_init__(self):
        if self.environment is None:
            self.environment = _environment.Environment()

        # set SRUN_CPUS_PER_TASK to CPUsPerNode if not explicitly set
        self.environment = self.environment.set_srun_cpus_per_task_if_unset(
            self.resources
        )

    @classmethod
    def from_filepath(
        cls, filepath: pathlib.Path, connection_id: t.Optional[uuid.UUID] = None
    ) -> "Config":
        """Initialize from a given file."""
        return cls._from_dict(
            read.read_config(filepath), connection_id=connection_id
        )

    @property
    def files_to_exclude(self) -> t.List[str]:
        return self.exclude or []

    def _to_dict(self) -> t.Dict:
        environment = (
            self.environment.to_dict() if self.environment is not None else {}
        )
        return {
            "Project": self.project,
            "Resources": self.resources,
            # Write stderr/stdout to given file to allow access to logs
            "Stdout": APPLICATION_LOGS_FILE,
            "Stderr": APPLICATION_LOGS_FILE,
            **environment,
        }

    def to_job_description(self, arguments: str) -> t.Dict:
        """Convert to UNICORE job description."""
        environment = (
            self.environment.to_job_description(arguments)
            if self.environment is not None
            else {}
        )
        return {
            **self.to_dict(),
            **environment,
        }

    def execution_environment_given(self) -> bool:
        return (
            self.environment is not None and self.environment.execution_given()
        )

    def add_env_vars(self, env_vars: t.Dict) -> None:
        self.environment.add_env_vars(env_vars)



import mlflow.entities as entities
import pytest

import mantik.testing.pyunicore as test_unicore
import mantik_compute_backend.unicore._properties as _properties


class TestSubmittedUnicoreRun:
    def test_run_id(self, example_job_properties):
        run = test_unicore._create_run(properties=example_job_properties)
        expected = "test-job"

        result = run.run_id

        assert result == expected

    @pytest.mark.parametrize(
        ("will_be_successful", "expected"),
        [
            (
                True,
                True,
            ),
            (
                False,
                False,
            ),
        ],
    )
    def test_wait(self, will_be_successful, expected, example_job_properties):
        run = test_unicore._create_run(
            will_be_successful=will_be_successful,
            properties=example_job_properties,
        )
        result = run.wait()

        assert result == expected

    @pytest.mark.parametrize(
        ("status", "expected"),
        [
            (
                _properties.Status.STAGING_IN,
                entities.RunStatus.SCHEDULED,
            ),
            (
                _properties.Status.READY,
                entities.RunStatus.SCHEDULED,
            ),
            (
                _properties.Status.QUEUED,
                entities.RunStatus.SCHEDULED,
            ),
            (
                _properties.Status.RUNNING,
                entities.RunStatus.RUNNING,
            ),
            (
                _properties.Status.STAGING_OUT,
                entities.RunStatus.RUNNING,
            ),
            (
                _properties.Status.SUCCESSFUL,
                entities.RunStatus.FINISHED,
            ),
            (
                _properties.Status.FAILED,
                entities.RunStatus.FAILED,
            ),
            (
                _properties.Status.UNKNOWN,
                entities.RunStatus.RUNNING,
            ),
        ],
    )
    def test_get_status(self, status, expected, example_job_properties):
        run = test_unicore._create_run(
            status=status, properties=example_job_properties
        )
        result = run.get_status()

        assert result == expected

    def test_cancel(self, example_job_properties):
        run = test_unicore._create_run(properties=example_job_properties)

        run.cancel()



import uuid

import pytest

import mantik.runs.schemas as run_schemas
import mantik.utils.mantik_api.code_repository as code_api
import mantik.utils.mantik_api.experiment_repository as experiment_api


@pytest.fixture(scope="session")
def sample_experiment_repository_id():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_code_repository_id():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_project_id() -> uuid.UUID:
    return uuid.uuid4()


@pytest.fixture()
def sample_experiment_repository(sample_experiment_repository_id):
    return experiment_api.ExperimentRepository(
        experiment_repository_id=sample_experiment_repository_id,
        mlflow_experiment_id="123",
        name="Name",
        artifact_location="somewhere.com",
    )


@pytest.fixture()
def sample_code_repository(sample_code_repository_id):
    return code_api.CodeRepository(
        code_repository_id=sample_code_repository_id,
        code_repository_name="Name",
        uri="some/uri.git",
        access_token="1234",
    )


@pytest.fixture(scope="session")
def sample_run_configuration(
    sample_experiment_repository_id, sample_code_repository_id
) -> run_schemas.RunConfiguration:
    return run_schemas.RunConfiguration(
        name="Sample",
        experiment_repository_id=sample_experiment_repository_id,
        code_repository_id=sample_code_repository_id,
        branch="branch",
        commit="commit",
        data_repository_id=uuid.uuid4(),
        mlflow_mlproject_file_path="some/path/MLProject",
        entry_point="main",
        mlflow_parameters={"output": "hello world"},
        backend_config={},
    )


@pytest.fixture()
def fake_token():
    return "1234"



import dataclasses
import json
import logging
import pathlib
import typing as t
import uuid

import requests

import mantik.cli as cli
import mantik.compute_backend.config.core as core
import mantik.compute_backend.config.validate as validate
import mantik.tracking as tracking
import mantik.unicore.credentials as _unicore_credentials
import mantik.utils as utils
import mantik.utils.unicore.zip as unicore_zip

logger = logging.getLogger(__name__)

API_SUBMIT_PATH = "/submit"
_DEFAULT_COMPUTE_BACKEND_API_PATH = "/compute-backend"
_DEFAULT_COMPUTE_BACKEND_SUBDOMAIN = "computebackend"
_COMPUTE_BACKEND_SUBDOMAIN = utils.env.get_optional_env_var(
    "COMPUTE_BACKEND_SUB_DOMAIN", default=_DEFAULT_COMPUTE_BACKEND_SUBDOMAIN
)


@dataclasses.dataclass
class ComputeBackendClient:
    """Client for the compute backend.

    Parameters
    ----------
    compute_backend_url : str
        URL of the compute backend API.
    unicore_username : str
        UNICORE username.
    unicore_password : str
        Password for the respective UNICORE user.
    compute_budget_account : str
        Project to use for the compute budget.
    mlflow_tracking_uri : str
        URI of the MLflow tracking server.
    mlflow_tracking_token : str
        Authentication token for the MLflow tracking server.
    api_path : str, default="/compute-backend"
        API path of the compute backend.
        E.g. if `compute_backend_url="https://<host>.com"`, but the actual API
        listens at `https://<host>.com/any/path`, this path can be given here,
        i.e. `api_path=/any/path`.

    """

    compute_backend_url: str
    unicore_username: str
    unicore_password: str
    compute_budget_account: str
    mlflow_tracking_uri: str
    mlflow_tracking_token: str
    api_path: str = _DEFAULT_COMPUTE_BACKEND_API_PATH

    @classmethod
    def from_env(
        cls,
        api_path: str = _DEFAULT_COMPUTE_BACKEND_API_PATH,
        connection_id: t.Optional[uuid.UUID] = None,
    ) -> "ComputeBackendClient":
        """Create from environment variables."""
        unicore_credentials = (
            _unicore_credentials.UnicoreCredentials.get_credentials(
                connection_id
            )
        )
        compute_budget_account = utils.env.get_required_env_var(
            core._PROJECT_ENV_VAR
        )
        mlflow_tracking_uri = utils.env.get_required_env_var(
            utils.mlflow.TRACKING_URI_ENV_VAR
        )
        environment = tracking.track.init_tracking()
        return cls(
            compute_backend_url=mlflow_tracking_uri,
            unicore_username=unicore_credentials.username,
            unicore_password=unicore_credentials.password,
            compute_budget_account=compute_budget_account,
            mlflow_tracking_uri=mlflow_tracking_uri,
            mlflow_tracking_token=environment.token,
            api_path=api_path,
        )

    def __post_init__(self):
        """Ensure Compute Backend URL contains the correct subdomain."""
        with_path_prefix = f"{self.compute_backend_url}/{self.api_path}"
        with_sub_domain = utils.urls.replace_first_subdomain(
            with_path_prefix, replace_with=_COMPUTE_BACKEND_SUBDOMAIN
        )
        with_https = utils.urls.remove_double_slashes_from_path(with_sub_domain)
        self.compute_backend_url = with_https
        logger.debug("Client initialized")

    @property
    def url(self) -> str:
        """Return the API URL of the compute backend."""
        return self.compute_backend_url

    @property
    def submit_url(self) -> str:
        """Return the API URL for the submit endpoint."""
        return f"{self.url}{API_SUBMIT_PATH}"

    def submit_run(
        self,
        run_name: str,
        experiment_id: int,
        mlproject_path: t.Union[pathlib.Path, str],
        mlflow_parameters: t.Dict,
        backend_config: t.Union[pathlib.Path, str],
        entry_point: t.Optional[str] = None,
    ) -> requests.Response:
        """Submit a run.

        Parameters
        ----------
        run_name: str
            Name of the run
        experiment_id : int
            ID of the tracking experiment.
        mlproject_path : pathlib.Path
            Path to the MLproject directory.
        mlflow_parameters : dict
            Parameters to pass to the MLproject.
        backend_config : pathlib.Path
            Path within the MLproject directory to the backend config.
            E.g. `backend-config.json`.
        entry_point : str, default="main"
            Name of the entry point to execute.

        Returns
        -------
        requests.Response
            The response of the compute backend.

        """
        if entry_point is None:
            entry_point = "main"
        logger.debug(
            "Submitting MLproject %s for experiment %s",
            mlproject_path,
            experiment_id,
        )
        project_validator = validate.ProjectValidator(
            mlproject_path=mlproject_path,
            config_path=backend_config,
            mlflow_parameters=mlflow_parameters,
            entry_point=entry_point,
            logger_level=logger.level,
        )
        project_validator.validate()
        url = f"{self.submit_url}/{experiment_id}"
        data = self._generate_request_data(
            run_name=run_name,
            entry_point=entry_point,
            mlflow_parameters=mlflow_parameters,
            compute_backend_config=project_validator.config_relative_path.as_posix(),  # noqa
        )
        files = _generate_request_files(
            mlproject_path=project_validator.mlproject_path,
            config=project_validator.config,
        )
        headers = self._generate_headers()
        logger.debug(
            "Sending request to compute backend %s with headers %s and data %s",
            url,
            headers,
            data,
        )
        try:
            response = requests.post(
                url=url,
                headers=headers,
                data=data,
                files=files,
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.exception(
                "Job submission failed (%s): %s", e.response, e.response.content
            )
            if e.response.status_code == 413:
                message = (
                    "The files you submitted were too large. "
                    "Please consider transferring large files "
                    "(such as image files) manually using the "
                    "Mantik Remote File Service CLI. "
                    "(See mantik "
                    f"{cli.remote_file_service.unicore_file_service.GROUP_NAME} "  # noqa: E501
                    "--help). "
                    "You will also have to change the backend "
                    "configuration to use a remote image. "
                    "Then you can try to re-submit the job."
                )
            else:
                try:
                    message = e.response.json()["message"]
                except (KeyError, json.decoder.JSONDecodeError) as e:
                    message = e.response.text
            raise requests.HTTPError(f"Job submission failed: {message}") from e
        else:
            (
                experiment_id,
                run_id,
                unicore_job_id,
            ) = _extract_job_info_from_response_json(response.json())
            logger.debug(
                "Job submitted successful with experiment_id=%s,"
                " run_id=%s, unicore_job_id=%s.",
                experiment_id,
                run_id,
                unicore_job_id,
            )
            return response

    def _generate_headers(self):
        return {
            "Authorization": f"Bearer {self.mlflow_tracking_token}",
            "Accept": "application/json",
        }

    def _generate_request_data(
        self,
        run_name: str,
        entry_point: str,
        mlflow_parameters: t.Dict,
        compute_backend_config: str,
    ) -> t.Dict:
        return {
            "run_name": run_name,
            "entry_point": entry_point,
            "mlflow_parameters": json.dumps(mlflow_parameters),
            "unicore_user": self.unicore_username,
            "unicore_password": self.unicore_password,
            "compute_budget_account": self.compute_budget_account,
            "compute_backend_config": compute_backend_config,
            "mlflow_tracking_uri": self.mlflow_tracking_uri,
            "mlflow_tracking_token": self.mlflow_tracking_token,
        }


def _generate_request_files(
    mlproject_path: pathlib.Path, config: core.Config
) -> t.Dict:
    logger.debug("Zipping MLproject directory %s", mlproject_path)
    zipped = unicore_zip.zip_directory_with_exclusion(mlproject_path, config)
    return {"mlproject_zip": zipped.read()}


def _extract_job_info_from_response_json(
    response_json: t.Dict,
) -> t.Tuple[str, str, str]:
    experiment_id = response_json.get("experiment_id")
    run_id = response_json.get("run_id")
    unicore_job_id = response_json.get("unicore_job_id")
    return experiment_id, run_id, unicore_job_id



import functools
import json
import pathlib

import mlflow.projects as projects
import pyunicore.client

import mantik.compute_backend.config.core as core
import mantik.testing as testing
import mantik.unicore.credentials as unicore_credentials
import mantik.utils as mantik_utils
import mantik_compute_backend.unicore._client as _client
import mantik_compute_backend.unicore.backend as backend
import mantik_compute_backend.utils as utils

FILE_PATH = pathlib.Path(__file__).parent

ALL_ENV_VARS = [
    unicore_credentials._USERNAME_ENV_VAR,
    unicore_credentials._PASSWORD_ENV_VAR,
    core._PROJECT_ENV_VAR,
]


class FakeProject:
    pass


class TestUnicoreBackend:
    def test_run(
        self, monkeypatch, example_project_path, tmp_path, env_vars_set, caplog
    ):
        monkeypatch.setattr(
            pyunicore.client,
            "Transport",
            testing.pyunicore.FakeTransport,
        )
        fake_client_with_successful_login = functools.partial(
            testing.pyunicore.FakeClient,
            login_successful=True,
        )
        monkeypatch.setattr(
            pyunicore.client,
            "Client",
            fake_client_with_successful_login,
        )
        backend_config_path = (
            example_project_path / "compute-backend-config.json"
        )
        with open(backend_config_path) as infile:
            backend_config = json.load(infile)
        # Following env vars are set by MLflow before running a project.
        backend_config[
            projects.PROJECT_STORAGE_DIR
        ] = example_project_path.as_posix()

        # Point MLFLOW_TRACKING_URI to a temporary directory
        tracking_uri = tmp_path / "mlruns"

        run_name = "my incredible run"

        with env_vars_set(
            {
                # Following env vars must be set for the config.
                **{key: "test-val" for key in ALL_ENV_VARS},
                mantik_utils.mlflow.TRACKING_URI_ENV_VAR: tracking_uri.as_uri(),
                utils.mlflow.RUN_NAME_ENV_VAR: run_name,
            }
        ):
            submitted_run = backend.UnicoreBackend().run(
                project_uri=example_project_path.as_posix(),
                entry_point="main",
                params={"print": "test"},
                version=None,
                backend_config=backend_config,
                tracking_uri=None,
                experiment_id=None,
            )

            assert submitted_run._job._job.started
            assert "with name my incredible run" in caplog.text


def test_submit_job_in_staging_in_and_upload_input_files(
    example_project_path, example_project, example_config
):
    client = testing.pyunicore.FakeClient()
    entry_point = example_project.get_entry_point("main")
    parameters = {"print": "test"}
    apptainer_image = example_project_path / "mantik-test.sif"
    client = _client.Client(client)
    backend._submit_job_in_staging_in_and_upload_input_files(
        client=client,
        entry_point=entry_point,
        parameters=parameters,
        storage_dir="",
        input_files=[apptainer_image],
        config=example_config,
        run_id="example_run_id",
    )


def test_create_job_description_apptainer(
    example_project_path, example_project, example_config
):
    entry_point = example_project.get_entry_point("main")
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"
    expected = {
        "Executable": (
            "srun apptainer run mantik-test.sif python main.py whatever"
        ),
        "Arguments": [],
        "Project": "test-project",
        "Resources": {"Queue": "batch"},
        "RunUserPrecommandOnLoginNode": True,
        "RunUserPostcommandOnLoginNode": True,
        "Stderr": "mantik.log",
        "Stdout": "mantik.log",
    }
    result = backend.create_job_description(
        entry_point=entry_point,
        parameters=parameters,
        storage_dir=storage_dir,
        config=example_config,
    )

    # Environment contains additional MLFLOW env vars,
    # which depend on the execution environment
    expected_environment = {"SRUN_CPUS_PER_TASK": 100}
    actual_environment = result.pop("Environment")
    assert all(
        actual_environment[key] == value
        for key, value in expected_environment.items()
    )

    assert result == expected


def test_create_job_description_apptainer_entry_point_multi_line(
    example_project_path, example_project, example_config
):
    entry_point = example_project.get_entry_point("multi-line")
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"
    expected = {
        "Executable": (
            "srun apptainer run mantik-test.sif python main.py whatever  "
            "--o option1  --i option2"
        ),
        "Arguments": [],
        "Project": "test-project",
        "Resources": {"Queue": "batch"},
        "RunUserPrecommandOnLoginNode": True,
        "RunUserPostcommandOnLoginNode": True,
        "Stderr": "mantik.log",
        "Stdout": "mantik.log",
    }
    result = backend.create_job_description(
        entry_point=entry_point,
        parameters=parameters,
        storage_dir=storage_dir,
        config=example_config,
    )

    # Environment contains additional MLFLOW env vars,
    # which depend on the execution environment
    expected_environment = {"SRUN_CPUS_PER_TASK": 100}
    actual_environment = result.pop("Environment")
    assert all(
        actual_environment[key] == value
        for key, value in expected_environment.items()
    )

    assert result == expected


def test_create_job_description_python(
    example_project_path, example_project, example_config_for_python
):
    entry_point = example_project.get_entry_point("main")
    parameters = {"print": "whatever"}
    storage_dir = "test-dir"
    expected = {
        "Executable": (
            "precommand compute node && "
            "source /venv/bin/activate && python main.py whatever && "
            "postcommand compute node"
        ),
        "Arguments": [],
        "Project": "test-project",
        "Resources": {"Queue": "batch"},
        "RunUserPrecommandOnLoginNode": True,
        "RunUserPostcommandOnLoginNode": True,
        "Stderr": "mantik.log",
        "Stdout": "mantik.log",
    }
    result = backend.create_job_description(
        entry_point=entry_point,
        parameters=parameters,
        storage_dir=storage_dir,
        config=example_config_for_python,
    )

    # Removing Environment for assertion since it contains additional MLFLOW
    # env vars, which depend on the execution environment
    result.pop("Environment", None)

    assert result == expected



import mantik.cli
import mantik.remote_file_service
import mantik.unicore
import mantik.utils
from mantik.compute_backend.client import ComputeBackendClient
from mantik.tracking.track import init_tracking



import uuid

import pytest
import requests

import mantik.utils.mantik_api.models as models


@pytest.fixture(scope="session")
def model_id() -> uuid.UUID:
    return uuid.uuid4()


@pytest.fixture(scope="session")
def project_id() -> uuid.UUID:
    return uuid.uuid4()


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_model(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    sample_model_schema = {
        "modelId": str(model_id),
        "uri": "string",
        "location": "string",
        "connectionId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "mlflowParameters": "{}",
        "modelRepository": "{}",
    }

    with mock_mantik_api_request(
        method="GET",
        end_point=f"/mantik-api/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}",
        status_code=status_code,
        json_response=sample_model_schema,
        expected_error=expected,
    ) as (m, error):
        retrieved_model = models.get_one(
            project_id=project_id, model_id=model_id, token="test_token"
        )
        assert retrieved_model == models.GetModel.from_json(sample_model_schema)
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_get_all_project_models(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    sample_model_schema = {
        "modelId": str(model_id),
        "uri": "string",
        "location": "string",
        "connectionId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "mlflowParameters": "{}",
        "modelRepository": "{}",
    }

    with mock_mantik_api_request(
        method="GET",
        end_point=f"/mantik-api/projects/{str(project_id)}/models/trained",
        status_code=status_code,
        json_response={
            "totalRecords": 1,
            "models": [sample_model_schema],
        },
        expected_error=expected,
    ) as (m, error):
        retrieved_model = models.get_all(
            project_id=project_id, token="test_token"
        )
        assert retrieved_model == [
            models.GetModel.from_json(sample_model_schema)
        ]
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(204, None), (404, requests.HTTPError())]
)
def test_delete_model_entry(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    with mock_mantik_api_request(
        method="DELETE",
        end_point=f"/mantik-api/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}",
        status_code=status_code,
        json_response={},
        expected_error=expected,
    ) as (m, error):
        models.delete(
            project_id=project_id,
            model_id=model_id,
            token="test_token",
        )
        assert any(
            f"Model with ID: {model_id} has been deleted" in message
            for message in info_caplog.messages
        )
    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )


def test_create_model_entry(
    mock_mantik_api_request,
    add_data_repository,
    info_caplog,
    project_id,
    model_id,
):
    new_model_schema = models.PostPutModel(
        uri="s3://bucket/item.type",
        location="S3",
        connection_id=uuid.uuid4(),
        mlflow_parameters={},
        model_repository_id=None,
    )

    with mock_mantik_api_request(
        method="POST",
        end_point=f"/mantik-api/projects/{str(project_id)}/models/trained",
        status_code=201,
        json_response={"modelId": str(model_id)},
        expected_error=None,
    ):
        new_model_id = models.add(
            new_model_schema=new_model_schema,
            project_id=project_id,
            token="test_token",
        )
        assert new_model_id == model_id

        assert any(
            f"A model with ID: {model_id} has been created" in message
            for message in info_caplog.messages
        )


@pytest.mark.parametrize(
    ("status_code", "expected"), [(200, None), (404, requests.HTTPError())]
)
def test_update_model(
    mock_mantik_api_request,
    info_caplog,
    status_code,
    expected,
    model_id,
    project_id,
):
    updated_model_schema = models.PostPutModel(
        uri="s3://bucket/item.type",
        location="S3",
        connection_id=uuid.uuid4(),
        mlflow_parameters={},
        model_repository_id=None,
    )

    with mock_mantik_api_request(
        method="PUT",
        end_point=f"/mantik-api/projects/{str(project_id)}/"
        f"models/trained/{str(model_id)}",
        status_code=status_code,
        json_response={},
        expected_error=expected,
    ) as (m, error):
        models.update(
            project_id=project_id,
            model_id=model_id,
            updated_model_schema=updated_model_schema,
            token="test_token",
        )
        assert f"Model with ID: {str(model_id)} has been updated"

    if error:
        assert any(
            "Call to Mantik API" in message for message in info_caplog.messages
        )



import os
import unittest.mock
import uuid

import mlflow.exceptions
import pytest

import mantik.runs.local
import mantik.utils.mantik_api.code_repository


@pytest.fixture()
def mock_get_code(mocker, sample_code_repository):
    mocker.patch(
        "mantik.utils.mantik_api.code_repository.get_one",
        return_value=sample_code_repository,
    )
    yield


@pytest.fixture()
def mock_get_experiment(mocker, sample_experiment_repository):
    mocker.patch(
        "mantik.utils.mantik_api.experiment_repository.get_one",
        return_value=sample_experiment_repository,
    )
    yield


@pytest.fixture()
def mock_get_unique_run_name(mocker, sample_run_configuration):
    mocker.patch(
        "mantik.utils.mantik_api.experiment_repository.get_unique_run_name",
        return_value=sample_run_configuration.name,
    )
    yield


@pytest.fixture()
def mock_update_run_status():
    with unittest.mock.patch(
        "mantik.utils.mantik_api.run.update_run_status",
    ) as patch:
        yield patch


@pytest.fixture()
def mock_keyboard_interrupt(mocker):
    mocker.patch("mlflow.run", side_effect=KeyboardInterrupt)
    yield


@pytest.fixture()
def mock_mlflow_execution_exception(mocker):
    mocker.patch(
        "mlflow.run",
        side_effect=mlflow.exceptions.ExecutionException(message="test"),
    )
    yield


@pytest.fixture()
def mock_mlflow_run(mocker):
    class Run:
        def get_status(self):
            return "FINISHED"

    mocker.patch("mlflow.run", return_value=Run())
    yield


@pytest.fixture()
def set_fake_mlflow_tracking():
    os.environ[mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR] = "FAKE TOKEN"
    os.environ[mantik.utils.mlflow.TRACKING_URI_ENV_VAR] = "FAKE URL"
    yield
    os.unsetenv(mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR)
    os.unsetenv(mantik.utils.mlflow.TRACKING_URI_ENV_VAR)


def test_local_run(
    mock_get_code,
    mock_get_experiment,
    mock_get_unique_run_name,
    sample_run_configuration,
    sample_project_id,
    fake_token,
) -> None:
    mocked_local_run_manager = unittest.mock.Mock(
        spec=mantik.runs.local.LocalRunManager
    )

    mantik.runs.local.run(
        data=sample_run_configuration,
        project_id=sample_project_id,
        mantik_token=fake_token,
        run_manager=mocked_local_run_manager,
    )

    mocked_local_run_manager.clone_git_repo.assert_called()
    mocked_local_run_manager.start_local_run.assert_called()


def test_path_directory_of_mlproject_file() -> None:
    assert (
        mantik.runs.local.path_directory_of_mlproject_file(
            "123/mlproject/MLProject"
        )
        == "123/mlproject"
    )


def test_submit_local_run_and_keyboard_interrupt(
    mock_mantik_api_request,
    sample_run_configuration,
    tmpdir,
    mock_update_run_status,
    mock_keyboard_interrupt,
    set_fake_mlflow_tracking,
) -> None:
    mocked_local_run_manager = mantik.runs.local.LocalRunManager()
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()

    with pytest.raises(KeyboardInterrupt):
        with mock_mantik_api_request(
            method="POST",
            end_point=f"/mantik-api/projects/"
            f"{str(project_id)}/runs?submit=False",
            status_code=201,
            json_response={"runId": str(run_id)},
        ):
            mocked_local_run_manager.start_local_run(
                mlflow_experiment_id=0,
                data=sample_run_configuration,
                mantik_token="access",
                project_id=project_id,
                uri=tmpdir,
            )
    mlflow.end_run()
    assert mock_update_run_status.call_count == 2
    mock_update_run_status.assert_called_with(
        project_id=project_id, run_id=run_id, token="access", status="KILLED"
    )


def test_submit_local_run_and_execution_exception(
    mock_mantik_api_request,
    sample_run_configuration,
    tmpdir,
    mock_update_run_status,
    mock_mlflow_execution_exception,
    set_fake_mlflow_tracking,
) -> None:
    mocked_local_run_manager = mantik.runs.local.LocalRunManager()
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    with pytest.raises(mlflow.exceptions.ExecutionException):
        with mock_mantik_api_request(
            method="POST",
            end_point=f"/mantik-api/projects/"
            f"{str(project_id)}/runs?submit=False",
            status_code=201,
            json_response={"runId": str(run_id)},
        ):
            mocked_local_run_manager.start_local_run(
                mlflow_experiment_id=0,
                data=sample_run_configuration,
                mantik_token="access",
                project_id=project_id,
                uri=tmpdir,
            )
    mlflow.end_run()
    assert mock_update_run_status.call_count == 2
    mock_update_run_status.assert_called_with(
        project_id=project_id, run_id=run_id, token="access", status="FAILED"
    )


def test_submit_local_run(
    mock_mantik_api_request,
    sample_run_configuration,
    tmpdir,
    mock_update_run_status,
    mock_mlflow_run,
    set_fake_mlflow_tracking,
) -> None:
    mocked_local_run_manager = mantik.runs.local.LocalRunManager()
    project_id = uuid.uuid4()
    run_id = uuid.uuid4()
    with mock_mantik_api_request(
        method="POST",
        end_point=f"/mantik-api/projects/"
        f"{str(project_id)}/runs?submit=False",
        status_code=201,
        json_response={"runId": str(run_id)},
    ):
        mocked_local_run_manager.start_local_run(
            mlflow_experiment_id=0,
            data=sample_run_configuration,
            mantik_token="access",
            project_id=project_id,
            uri=tmpdir,
        )
    mlflow.end_run()
    assert mock_update_run_status.call_count == 2
    mock_update_run_status.assert_called_with(
        project_id=project_id, run_id=run_id, token="access", status="FINISHED"
    )



import json
import typing as t
import uuid

import click

import mantik.authentication.auth
import mantik.cli.main as main
import mantik.utils.env
import mantik.utils.mantik_api.client
import mantik.utils.mantik_api.models

GROUP_NAME = "models"
_PROJECT_ID_ENV_VAR = "MANTIK_PROJECT_ID"
_MODEL_ID_ENV_VAR = "MANTIK_MODEL_ID"


def _access_token_from_env_vars() -> str:
    return mantik.authentication.auth.get_valid_access_token()


@main.cli.group(GROUP_NAME)
def cli() -> None:
    """Interaction with models through the mantik api."""


def mlflow_parameters_decoder(ctx, param, value) -> t.Optional[dict]:
    try:
        return json.loads(value) if value is not None else None
    except json.decoder.JSONDecodeError:
        raise TypeError(
            """MLflow parameters are in the wrong format. Try: '{"key": "value"}'"""  # noqa
        )


@cli.command("list")
@click.option(
    "--project-id",
    type=uuid.UUID,
    required=True,
    envvar=_PROJECT_ID_ENV_VAR,
)
def get_all_project_models(project_id: uuid.UUID) -> None:
    """Print details of all trained models in project."""
    models = mantik.utils.mantik_api.models.get_all(
        project_id=project_id,
        token=_access_token_from_env_vars(),
    )
    click.echo(models)


@cli.command("get-one")
@click.option(
    "--project-id",
    type=uuid.UUID,
    required=True,
    envvar=_PROJECT_ID_ENV_VAR,
)
@click.option(
    "--model-id", type=uuid.UUID, required=True, envvar=_MODEL_ID_ENV_VAR
)
def get_model(project_id: uuid.UUID, model_id: uuid.UUID) -> None:
    """Print details of specific trained model."""

    model = mantik.utils.mantik_api.models.get_one(
        project_id=project_id,
        model_id=model_id,
        token=_access_token_from_env_vars(),
    )
    click.echo(model)


@cli.command("delete")
@click.option(
    "--project-id",
    type=uuid.UUID,
    required=True,
    envvar=_PROJECT_ID_ENV_VAR,
)
@click.option(
    "--model-id", type=uuid.UUID, required=True, envvar=_MODEL_ID_ENV_VAR
)
def delete_model(project_id: uuid.UUID, model_id: uuid.UUID) -> None:
    """Delete a trained model."""
    mantik.utils.mantik_api.models.delete(
        project_id=project_id,
        model_id=model_id,
        token=_access_token_from_env_vars(),
    )


@cli.command("add")
@click.option(
    "--project-id",
    type=uuid.UUID,
    required=True,
    envvar=_PROJECT_ID_ENV_VAR,
)
@click.option("--uri", type=str, required=True)
@click.option("--location", type=str, required=True)
@click.option("--connection-id", type=uuid.UUID, required=False)
@click.option(
    "--mlflow-parameters",
    type=str,
    required=False,
    help="""JSON data as a string.""",
    callback=mlflow_parameters_decoder,
)
@click.option("--model-repository-id", type=uuid.UUID, required=False)
def create_model_entry(
    project_id: uuid.UUID,
    uri: str,
    location: str,
    connection_id: t.Optional[uuid.UUID],
    mlflow_parameters: t.Optional[dict],
    model_repository_id: t.Optional[uuid.UUID],
) -> None:
    """Add a new trained model entry."""

    mantik.utils.mantik_api.models.add(
        project_id=project_id,
        new_model_schema=mantik.utils.mantik_api.models.PostPutModel(
            uri=uri,
            location=location,
            connection_id=connection_id,
            mlflow_parameters=mlflow_parameters,
            model_repository_id=model_repository_id,
        ),
        token=_access_token_from_env_vars(),
    )


@cli.command("update")
@click.option(
    "--project-id",
    type=uuid.UUID,
    required=True,
    envvar=_PROJECT_ID_ENV_VAR,
)
@click.option(
    "--model-id", type=uuid.UUID, required=True, envvar=_MODEL_ID_ENV_VAR
)
@click.option("--uri", type=str, required=True)
@click.option("--location", type=str, required=True)
@click.option("--connection-id", type=uuid.UUID, required=False)
@click.option(
    "--mlflow-parameters",
    type=str,
    required=False,
    help="""JSON data as a string.""",
    callback=mlflow_parameters_decoder,
)
@click.option("--model-repository-id", type=uuid.UUID, required=False)
def update_model_entry(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    uri: str,
    location: str,
    connection_id: t.Optional[uuid.UUID],
    mlflow_parameters: t.Optional[dict],
    model_repository_id: t.Optional[uuid.UUID],
) -> None:
    """Update details of existing trained model."""

    mantik.utils.mantik_api.models.update(
        project_id=project_id,
        model_id=model_id,
        updated_model_schema=mantik.utils.mantik_api.models.PostPutModel(
            uri=uri,
            location=location,
            connection_id=connection_id,
            mlflow_parameters=mlflow_parameters,
            model_repository_id=model_repository_id,
        ),
        token=_access_token_from_env_vars(),
    )



"""Interact with trained models in mantik api"""
import dataclasses
import typing as t
import uuid

from mantik.utils.mantik_api.client import logger
from mantik.utils.mantik_api.client import send_request_to_mantik_api

T = t.TypeVar("T")


def optional_cast(_type: T, optional_value: t.Optional[t.Any]) -> t.Optional[T]:
    if optional_value is None:
        return None

    return _type(optional_value)


@dataclasses.dataclass
class ModelBase:
    uri: str
    location: str
    connection_id: t.Optional[uuid.UUID]
    mlflow_parameters: t.Optional[dict]


@dataclasses.dataclass
class GetModel(ModelBase):
    model_id: uuid.UUID
    model_repository: dict

    @classmethod
    def from_json(cls, json_str):
        return cls(
            model_id=uuid.UUID(json_str["modelId"]),
            uri=json_str["uri"],
            location=json_str["location"],
            connection_id=optional_cast(uuid.UUID, json_str["connectionId"]),
            mlflow_parameters=json_str["mlflowParameters"],
            model_repository=json_str["modelRepository"],
        )


@dataclasses.dataclass
class PostPutModel(ModelBase):
    model_repository_id: t.Optional[uuid.UUID]

    def to_dict(self):
        return {
            "uri": self.uri,
            "location": self.location,
            "connectionId": optional_cast(str, self.connection_id),
            "mlflowParameters": self.mlflow_parameters,
            "modelRepositoryId": optional_cast(str, self.model_repository_id),
        }


def get_all(
    project_id: uuid.UUID,
    token: str,
) -> t.List[GetModel]:
    """Retrieves trained model entries through the Mantik API"""

    endpoint = f"/mantik-api/projects/{str(project_id)}/models/trained"
    response = send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return [GetModel.from_json(model) for model in response.json()["models"]]


def get_one(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
) -> GetModel:
    """Retrieves a trained model entry from the Mantik API"""

    endpoint = (
        f"/mantik-api/projects/{str(project_id)}/models/trained/{str(model_id)}"
    )
    response = send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return GetModel.from_json(response.json())


def delete(
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
):
    """Deletes a trained model entry through the Mantik API"""

    endpoint = (
        f"/mantik-api/projects/{str(project_id)}/models/trained/{str(model_id)}"
    )
    send_request_to_mantik_api(
        method="DELETE", data={}, url_endpoint=endpoint, token=token
    )
    logger.info(f"Model with ID: {model_id} has been deleted")


def add(
    new_model_schema: PostPutModel,
    project_id: uuid.UUID,
    token: str,
) -> uuid.UUID:
    """Creates a trained model entry through the Mantik API"""

    data = new_model_schema.to_dict()
    endpoint = f"/mantik-api/projects/{str(project_id)}/models/trained"
    response = send_request_to_mantik_api(
        method="POST", data=data, url_endpoint=endpoint, token=token
    )
    new_model_id = response.json()["modelId"]
    logger.info(f"A model with ID: {new_model_id} has been created")

    return uuid.UUID(new_model_id)


def update(
    updated_model_schema: PostPutModel,
    project_id: uuid.UUID,
    model_id: uuid.UUID,
    token: str,
):
    """Updates a trained model entry through the Mantik API"""

    data = updated_model_schema.to_dict()
    endpoint = (
        f"/mantik-api/projects/{str(project_id)}/models/trained/{str(model_id)}"
    )
    send_request_to_mantik_api(
        method="PUT", data=data, url_endpoint=endpoint, token=token
    )
    logger.info(f"Model with ID: {str(model_id)} has been updated")



import logging
import pathlib
import typing as t
import uuid

import click

import mantik
import mantik.cli._options as _main_options
import mantik.cli.runs.runs as runs
import mantik.cli.utils as cli_utils
import mantik.compute_backend.config.core as core
import mantik.compute_backend.config.read as read
import mantik.utils.env_vars as env_vars
import mantik.utils.mantik_api as mantik_api

logger = logging.getLogger(__name__)


@runs.cli.command("submit")
@click.argument(
    "mlproject-path",
    type=click.Path(path_type=pathlib.Path),
    required=True,
)
@click.option(
    "--run-name",
    default=None,
    type=str,
    help="Name of the Run.",
    required=True,
)
@click.option(
    "--entry-point",
    required=False,
    default="main",
    show_default=True,
    help="Entrypoint of the MLproject file.",
)
@click.option(
    "--backend-config",
    type=click.Path(dir_okay=False, path_type=pathlib.Path),
    required=True,
    help="Relative or absolute path to backend config file.",
)
@click.option(
    "--project-id",
    type=uuid.UUID,
    default=None,
    help=f"""Project ID on Mantik.

        If not specified, it is inferred from the environment variable
        {env_vars._PROJECT_ID_ENV_VAR}.

    """,
)
@click.option(
    "--data-repository-id",
    type=uuid.UUID,
    default=None,
    help=f"""Data repository ID on Mantik

        If not specified, it is inferred from the environment variable
        {env_vars._DATA_REPOSITORY_ID_ENV_VAR}.

    """,
)
@click.option(
    "--experiment-repository-id",
    type=uuid.UUID,
    default=None,
    help=f"""Experiment repository ID on Mantik

        If not specified, it is inferred from the environment variable
        {env_vars._EXPERIMENT_REPOSITORY_ID_ENV_VAR}.

    """,
)
@click.option(
    "--code-repository-id",
    type=uuid.UUID,
    default=None,
    help=f"""Code repository ID on Mantik

        If not specified, it is inferred from the environment variable
        {env_vars._CODE_REPOSITORY_ID_ENV_VAR}.

    """,
)
@click.option(
    "--branch",
    default=None,
    type=str,
    help="Name of the code repository's branch you want to use for this run",
    required=True,
)
@click.option(
    "--commit",
    default=None,
    type=str,
    help="""Name of the code repository's full commit hash you want to use for
        this run.

        If both branch and commit hash are given, the commit hash is preferred
        over the branch.
    """,
    required=False,
)
@click.option(
    "--compute-budget-account",
    type=str,
    default=None,
    help=f"""Name of your Compute Budget Account on HPC

        If not specified, it is inferred from the environment variable
        {core._PROJECT_ENV_VAR}.

    """,
)
@click.option(
    "--parameter", "-P", show_default=True, default=lambda: [], multiple=True
)
@_main_options.VERBOSE
@click.option(
    "--connection-id",
    required=True,
    default=None,
    type=uuid.UUID,
    help="Connection ID of your UNICORE Connection stored on Mantik. "
    "This is preferred over using env vars.",
)
def run_project(
    run_name: str,
    mlproject_path: pathlib.Path,
    entry_point: str,
    backend_config: pathlib.Path,
    parameter: t.List[str],
    verbose: bool,  # noqa
    project_id: t.Optional[uuid.UUID],
    data_repository_id: t.Optional[uuid.UUID],
    experiment_repository_id: t.Optional[uuid.UUID],
    code_repository_id: t.Optional[uuid.UUID],
    branch: t.Optional[str],
    commit: t.Optional[str],
    compute_budget_account: t.Optional[str],
    connection_id: t.Optional[uuid.UUID],
) -> None:
    """Submit an MLflow project as a run to the Mantik Compute Backend.

    Note that `MLPROJECT_PATH` is the relative path to the MLflow project
    folder with your Code Repository as root.

    Remember that when you submit a run, the code is retrieved from your
    remote Git Code Repository. So make sure to commit and push your
    changes before submitting a run! The only file read from your local
    system is the backend config.

    To find the respective required IDs make sure to check Mantik's UI

    """
    entry_point = entry_point or "main"
    project_id = project_id or mantik.utils.env.get_required_env_var(
        env_vars._PROJECT_ID_ENV_VAR
    )
    data_repository_id = (
        data_repository_id
        or mantik.utils.env.get_optional_env_var(
            env_vars._DATA_REPOSITORY_ID_ENV_VAR
        )
    )
    experiment_repository_id = (
        experiment_repository_id
        or mantik.utils.env.get_required_env_var(
            env_vars._EXPERIMENT_REPOSITORY_ID_ENV_VAR
        )
    )
    code_repository_id = (
        code_repository_id
        or mantik.utils.env.get_required_env_var(
            env_vars._CODE_REPOSITORY_ID_ENV_VAR
        )
    )
    compute_budget_account = (
        compute_budget_account
        or mantik.utils.env.get_required_env_var(core._PROJECT_ENV_VAR)
    )
    if branch is None and commit is None:
        raise ValueError(
            "Either provide a branch or full commit hash to submit"
        )
    elif branch is not None and commit is not None:
        logger.warning(
            "Both branch name %s and commit hash %s given, using commit hash to"
            " submit run",
            branch,
            commit,
        )

    logger.debug("Parsing MLflow entry point parameters")
    parameters = cli_utils.dict_from_list(parameter)

    config = read.read_config(backend_config)
    data = {
        "name": run_name,
        "experimentRepositoryId": str(experiment_repository_id),
        "codeRepositoryId": str(code_repository_id),
        "branch": branch,
        "commit": commit,
        "dataRepositoryId": str(data_repository_id)
        if data_repository_id is not None
        else None,
        "connectionId": str(connection_id),
        "computeBudgetAccount": compute_budget_account,
        "mlflowMlprojectFilePath": mlproject_path.as_posix(),
        "entryPoint": entry_point,
        "mlflowParameters": parameters,
        "backendConfig": config,
    }
    token = mantik.authentication.auth.get_valid_access_token()
    response = mantik_api.run.submit_run(
        project_id=project_id, submit_run_data=data, token=token
    )
    click.echo(response.content)



import typing as t
import uuid

import click

import mantik.cli.main as main
import mantik.cli.remote_file_service._arguments as _arguments
import mantik.cli.remote_file_service._options as _options
import mantik.remote_file_service.data_client as data_client

GROUP_NAME = "unicore-file-service"


@main.cli.group(GROUP_NAME)
def cli() -> None:
    """Interaction with the unicore file service."""


@cli.command("copy-file")
@_arguments.SOURCE
@_arguments.TARGET
@_options.CONNECTION_ID
@_options.DATA_REPOSITORY_NAME
@_options.DATA_REPOSITORY_DESCRIPTION
def copy_file(
    source: str,
    target: str,
    connection_id: t.Optional[uuid.UUID],
    name: t.Optional[str],
    description: t.Optional[str],
) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    file = uc_fs.copy_file(
        source=source,
        target=target,
        data_repository_name=name,
        data_repository_description=description,
    )
    click.echo(file)


@cli.command("remove-file")
@_arguments.TARGET
@_options.CONNECTION_ID
@_options.DATA_REPOSITORY_NAME
@_options.DATA_REPOSITORY_ID
def remove_file(
    target: str,
    connection_id: t.Optional[uuid.UUID],
    name: t.Optional[str],
    id: t.Optional[uuid.UUID],
) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    uc_fs.remove_file(
        target=target, data_repository_name=name, data_repository_id=id
    )
    click.echo("File removed!")


@cli.command("create-file")
@_arguments.TARGET
@_options.CONNECTION_ID
@_options.DATA_REPOSITORY_NAME
@_options.DATA_REPOSITORY_DESCRIPTION
def create_file_if_not_exists_file(
    target: str,
    connection_id: t.Optional[uuid.UUID],
    name: t.Optional[str],
    description: t.Optional[str],
) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    file = uc_fs.create_file_if_not_exists(
        target=target,
        data_repository_name=name,
        data_repository_description=description,
    )
    click.echo(file)


@cli.command("list-directory")
@_arguments.TARGET
@_options.CONNECTION_ID
def list_directory(target: str, connection_id: t.Optional[uuid.UUID]) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    directory = uc_fs.list_directory(target=target)
    click.echo(directory)


@cli.command("create-directory")
@_arguments.TARGET
@_options.CONNECTION_ID
@_options.DATA_REPOSITORY_NAME
@_options.DATA_REPOSITORY_DESCRIPTION
def create_directory(
    target: str,
    connection_id: t.Optional[uuid.UUID],
    name: t.Optional[str],
    description: t.Optional[str],
) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    directory = uc_fs.create_directory(
        target=target,
        data_repository_name=name,
        data_repository_description=description,
    )
    click.echo(directory)


@cli.command("remove-directory")
@_arguments.TARGET
@_options.CONNECTION_ID
@_options.DATA_REPOSITORY_NAME
@_options.DATA_REPOSITORY_ID
def remove_directory(
    target: str,
    connection_id: t.Optional[uuid.UUID],
    name: t.Optional[str],
    id: t.Optional[uuid.UUID],
) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    uc_fs.remove_directory(
        target=target, data_repository_name=name, data_repository_id=id
    )
    click.echo("Directory removed!")


@cli.command("copy-directory")
@_arguments.SOURCE
@_arguments.TARGET
@_options.CONNECTION_ID
@_options.DATA_REPOSITORY_NAME
@_options.DATA_REPOSITORY_DESCRIPTION
def copy_directory(
    source: str,
    target: str,
    connection_id: t.Optional[uuid.UUID],
    name: t.Optional[str],
    description: t.Optional[str],
) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    directory = uc_fs.copy_directory(
        source=source,
        target=target,
        data_repository_name=name,
        data_repository_description=description,
    )
    click.echo(directory)


@cli.command("exists")
@_arguments.TARGET
@_options.CONNECTION_ID
def exists(target: str, connection_id: t.Optional[uuid.UUID]) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    _exists = uc_fs.exists(target=target)
    click.echo(_exists)


@cli.command("user")
@_options.CONNECTION_ID
def user(connection_id: t.Optional[uuid.UUID]) -> None:
    unicore_file_service = _import_unicore_file_service()
    uc_fs = data_client.DataClient.from_env(
        connection_id=connection_id,
        remote_fs_type=unicore_file_service.UnicoreFileService,
    )
    click.echo(uc_fs.user)


def _import_unicore_file_service():
    try:
        import mantik.remote_file_service.unicore_file_service as unicore_file_service  # noqa: E501
    except ModuleNotFoundError as e:
        raise RuntimeError(
            "Extras for UNICORE file service must be "
            'installed via `pip install "mantik[unicore]"`'
        ) from e
    else:
        return unicore_file_service



import pathlib

import pytest

import mantik.compute_backend as compute_backend
import mantik.compute_backend.config.exceptions as exceptions
import mantik.testing as testing


class TestConfig:
    @pytest.mark.parametrize(
        ("env_vars", "d", "expected"),
        [
            # Test cases: environment variables not set.
            ([], {}, exceptions.ConfigValidationError()),
            (
                testing.config.ALL_ENV_VARS[:1],
                {},
                exceptions.ConfigValidationError(),
            ),
            (
                testing.config.ALL_ENV_VARS[:2],
                {},
                exceptions.ConfigValidationError(),
            ),
            (
                testing.config.ALL_ENV_VARS,
                {},
                exceptions.ConfigValidationError(),
            ),
            # Test case: backend config missing Resources section.
            (
                testing.config.ALL_ENV_VARS,
                {"Queue": "batch"},
                exceptions.ConfigValidationError(),
            ),
            # Test case: Everything correct.
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                    "Resources": {"Queue": "batch"},
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                    env=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                        ),
                    ),
                ),
            ),
            # Test case: No environment given
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                ),
            ),
            # Test case: No environment given, but resources
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {
                        "Queue": "batch",
                        "Runtime": "2h",
                        "Nodes": 1,
                        "TotalCPUs": 2,
                        "CPUsPerNode": 3,
                        "GPUsPerNode": 4,
                        "MemoryPerNode": "10G",
                        "Reservation": "test-reservation",
                        "NodeConstraints": "mem192",
                        "QoS": "test-qos",
                    },
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        runtime="2h",
                        nodes=1,
                        total_cpus=2,
                        cpus_per_node=3,
                        gpus_per_node=4,
                        memory_per_node="10G",
                        reservation="test-reservation",
                        node_constraints="mem192",
                        qos="test-qos",
                    ),
                ),
            ),
            # Test case: Only modules given in environment
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Environment": {
                        "Modules": ["module"],
                    },
                    "Resources": {"Queue": "batch"},
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                    env=compute_backend.config.environment.Environment(
                        modules=["module"],
                    ),
                ),
            ),
            # Test case: Local Apptainer image given.
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                        },
                    },
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                    env=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                        ),
                    ),
                ),
            ),
            # Test case: Local Apptainer image given with additional options
            # as list of strings
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                            "Options": ["--nv", "-B /data:/data"],
                        },
                    },
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                    env=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                            options=["--nv", "-B /data:/data"],
                        )
                    ),
                ),
            ),
            # Test case: Local Apptainer image given with additional options
            # as string
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "image.sif",
                            "Options": "--nv -B /data:/data",
                        },
                    },
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                    env=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("image.sif"),
                            options=["--nv -B /data:/data"],
                        )
                    ),
                ),
            ),
            # Test case: Remote Apptainer image given.
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "/absolute/path/to/image.sif",
                            "Type": "remote",
                        },
                    },
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                    env=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("/absolute/path/to/image.sif"),
                            type="remote",
                        ),
                    ),
                ),
            ),
            # Test case: More variables and modules given.
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "/absolute/path/to/image.sif",
                            "Type": "remote",
                        },
                        "Variables": {"TEST_ENV_VAR": "value"},
                        "Modules": [
                            "TensorFlow/2.5.0-Python-3.8.5",
                            "Horovod/0.23.0-Python-3.8.5",
                            "PyTorch/1.8.1-Python-3.8.5",
                        ],
                    },
                },
                testing.config._create_config(
                    resources=compute_backend.config.resources.Resources(
                        queue="batch"
                    ),
                    env=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("/absolute/path/to/image.sif"),
                            type="remote",
                        ),
                        modules=[
                            "TensorFlow/2.5.0-Python-3.8.5",
                            "Horovod/0.23.0-Python-3.8.5",
                            "PyTorch/1.8.1-Python-3.8.5",
                        ],
                        variables={"TEST_ENV_VAR": "value"},
                    ),
                ),
            ),
            # Test case: Remote Apptainer image given but relative path.
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch"},
                    "Environment": {
                        "Apptainer": {
                            "Path": "../relative/path/to/image.sif",
                            "Type": "remote",
                        },
                    },
                },
                compute_backend.config.exceptions.ConfigValidationError(),
            ),
            # Test case: Config entry has incorrect type.
            (
                testing.config.ALL_ENV_VARS,
                {
                    "UnicoreApiUrl": "test-url",
                    "Resources": {"Queue": "batch", "Nodes": "incorrect type"},
                },
                exceptions.ConfigValidationError(),
            ),
        ],
    )
    @pytest.mark.usefixtures("unset_tracking_token_env_var_before_execution")
    def test_from_dict(
        self, expect_raise_if_exception, env_vars_set, env_vars, d, expected
    ):
        env_vars = {
            var: testing.config.DEFAULT_ENV_VAR_VALUE for var in env_vars
        }

        with env_vars_set(env_vars), expect_raise_if_exception(expected):
            result = compute_backend.config.core.Config.from_dict(d)

            assert result == expected

    @pytest.mark.parametrize(
        ("config", "expected"),
        [
            # Test case: Only project, resources and
            # environment are included in the returned dict.
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("test-image")
                        ),
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {"Queue": "batch"},
                    "Arguments": [],
                    "RunUserPrecommandOnLoginNode": True,
                    "RunUserPostcommandOnLoginNode": True,
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                },
            ),
            # Test case: All given values included.
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        runtime="1h",
                        nodes=2,
                        total_cpus=48,
                        cpus_per_node=24,
                        memory_per_node="10G",
                        gpus_per_node=1,
                        reservation="test-reservation",
                        node_constraints="test-node-constraints",
                        qos="test-qos",
                    ),
                    environment=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("test-image"),
                        ),
                        variables={"TEST": "test"},
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {
                        "Queue": "batch",
                        "Runtime": "1h",
                        "Nodes": 2,
                        "TotalCPUs": 48,
                        "CPUsPerNode": 24,
                        "GPUS": 1,
                        "MemoryPerNode": "10G",
                        "Reservation": "test-reservation",
                        "NodeConstraints": "test-node-constraints",
                        "QoS": "test-qos",
                    },
                    "Environment": {"TEST": "test", "SRUN_CPUS_PER_TASK": "24"},
                    "Arguments": [],
                    "RunUserPrecommandOnLoginNode": True,
                    "RunUserPostcommandOnLoginNode": True,
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                },
            ),
        ],
    )
    def test_to_dict(self, config, expected):
        result = config.to_dict()

        assert result == expected

    @pytest.mark.parametrize(
        ("config", "expected"),
        [
            # Test case: No environment section, and cpus_per_node unset
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                    ),
                ),
                None,
            ),
            # Test case: Environment section, but no Variables, and
            # cpus_per_node unset
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=compute_backend.config.environment.Environment(),  # noqa E501
                ),
                None,
            ),
            # Test case: Environment section with Variables, and
            # cpus_per_node unset
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=compute_backend.config.environment.Environment(
                        variables={"TEST_VAR": "test_value"}
                    ),
                ),
                {"TEST_VAR": "test_value"},
            ),
            # Test case: SRUN_CPUS_PER_TASK set explicitly by user,
            # and cpus_per_node unset
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                    ),
                    environment=compute_backend.config.environment.Environment(
                        variables={"SRUN_CPUS_PER_TASK": "2"},
                    ),
                ),
                {"SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: No environment section, and cpus_per_node set
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=2,
                    ),
                ),
                {"SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: Environment section, but no Variables, and
            # cpus_per_node set
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=2,
                    ),
                    environment=compute_backend.config.environment.Environment(),  # noqa E501
                ),
                {"SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: Environment section with Variables, and
            # cpus_per_node set
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=2,
                    ),
                    environment=compute_backend.config.environment.Environment(
                        variables={"TEST_VAR": "test_value"},
                    ),
                ),
                {"TEST_VAR": "test_value", "SRUN_CPUS_PER_TASK": "2"},
            ),
            # Test case: SRUN_CPUS_PER_TASK set explicitly by user, and
            # cpus_per_node set
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=1,
                    ),
                    environment=compute_backend.config.environment.Environment(
                        variables={
                            "TEST_VAR": "test_value",
                            "SRUN_CPUS_PER_TASK": "2",
                        }
                    ),
                ),
                {"TEST_VAR": "test_value", "SRUN_CPUS_PER_TASK": "2"},
            ),
        ],
    )
    def test_with_optional_add_srun_cpus_per_task_to_environment(
        self, config, expected
    ) -> None:
        result = config.environment.variables

        assert result == expected

    @pytest.mark.parametrize(
        ("test_config", "expected"),
        [
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=1,
                    ),
                    environment=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Apptainer(
                            path=pathlib.Path("test-image"),
                        ),
                        variables={"TEST": "test"},
                        pre_run_command_on_compute_node="echo precommand on compute node",  # noqa E501
                        post_run_command_on_compute_node="echo postcommand on compute node",  # noqa E501
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {"Queue": "batch", "CPUsPerNode": 1},
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                    "Environment": {"TEST": "test", "SRUN_CPUS_PER_TASK": "1"},
                    "RunUserPrecommandOnLoginNode": True,
                    "Executable": (
                        "echo precommand on compute node && "
                        "srun apptainer run test-image "
                        "python main.py arg1 arg2 && "
                        "echo postcommand on compute node"
                    ),
                    "Arguments": [],
                    "RunUserPostcommandOnLoginNode": True,
                },
            ),
            (
                compute_backend.config.core.Config(
                    api_url="not_included",
                    user="not_included",
                    password="not_included",
                    project="test-project",
                    resources=compute_backend.config.resources.Resources(
                        queue="batch",
                        cpus_per_node=1,
                    ),
                    environment=compute_backend.config.environment.Environment(
                        execution=compute_backend.config.executable.Python(
                            path=pathlib.Path("/venv"),
                        ),
                        variables={"TEST": "test"},
                        pre_run_command_on_compute_node="echo precommand on compute node",  # noqa E501
                        post_run_command_on_compute_node="echo postcommand on compute node",  # noqa E501
                    ),
                ),
                {
                    "Project": "test-project",
                    "Resources": {"Queue": "batch", "CPUsPerNode": 1},
                    "Stdout": "mantik.log",
                    "Stderr": "mantik.log",
                    "Environment": {"TEST": "test", "SRUN_CPUS_PER_TASK": "1"},
                    "RunUserPrecommandOnLoginNode": True,
                    "Arguments": [],
                    "Executable": "echo precommand on compute node && source /venv/bin/activate && python main.py arg1 arg2 && echo postcommand on compute node",  # noqa E501
                    "RunUserPostcommandOnLoginNode": True,
                },
            ),
        ],
    )
    def test_to_job_description(self, test_config, expected):
        test_arguments_string = "python main.py arg1 arg2"
        result = test_config.to_job_description(test_arguments_string)
        assert result == expected



import os
import pathlib

import pytest

import mantik.compute_backend.config.environment as environment
import mantik.compute_backend.config.exceptions as exceptions
import mantik.compute_backend.config.executable as executable
import mantik.testing as testing


class TestEnvironment:
    @pytest.mark.parametrize(
        ("env_vars", "d", "expected"),
        [
            (
                testing.config.ALL_ENV_VARS,
                {
                    "Apptainer": {
                        "Path": "image.sif",
                    },
                },
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("image.sif")
                ),
            ),
            (
                testing.config.ALL_ENV_VARS,
                {"Python": "/path/to/venv"},
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
            ),
            (
                testing.config.ALL_ENV_VARS,
                {"Python": {"Path": "/path/to/venv"}},
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
            ),
            (
                testing.config.ALL_ENV_VARS,
                {
                    "Python": {"Path": "/path/to/venv"},
                    "PreRunCommandOnLoginNode": "precommand login node",
                    "PostRunCommandOnLoginNode": "postcommand login node",
                    "PreRunCommandOnComputeNode": "precommand compute node",
                    "PostRunCommandOnComputeNode": "postcommand compute node",
                },
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node="precommand login node",
                    post_run_command_on_login_node="postcommand login node",
                    pre_run_command_on_compute_node="precommand compute node",
                    post_run_command_on_compute_node="postcommand compute node",
                ),
            ),
        ],
    )
    def test_from_dict(
        self, monkeypatch, expect_raise_if_exception, env_vars, d, expected
    ):
        if "MLFLOW_TRACKING_TOKEN" in os.environ:
            del os.environ["MLFLOW_TRACKING_TOKEN"]
        if "MLFLOW_EXPERIMENT_ID" in os.environ:
            del os.environ["MLFLOW_EXPERIMENT_ID"]
        for key in env_vars:
            os.environ[key] = testing.config.DEFAULT_ENV_VAR_VALUE

        with expect_raise_if_exception(expected):
            result = environment.Environment.from_dict(d)

            assert result == expected

    @pytest.mark.parametrize(
        (
            "env",
            "expected",
        ),
        [
            # Test case: No execution environment
            (
                environment.Environment(),
                None,
            ),
            # Test case: Apptainer
            (
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
                "srun apptainer run venv",
            ),
            # Test case: Python
            (
                testing.config._create_python_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                ),
                "source /path/to/venv/bin/activate",
            ),
        ],
    )
    def test_create_execution_command(self, env, expected):
        result = env._create_execution_command()

        assert expected == result

    @pytest.mark.parametrize(
        (
            "env",
            "expected",
        ),
        [
            # Test case: Apptainer with no modules and no precommand
            (
                testing.config._create_apptainer_environment(
                    modules=None, path=pathlib.Path("/path/to/venv")
                ),
                None,
            ),
            # Test case: Apptainer with no modules and precommand
            (
                testing.config._create_apptainer_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node="a precommand",
                ),
                "a precommand",
            ),
            # Test case: Apptainer with 1 module and no precommand
            (
                testing.config._create_apptainer_environment(
                    modules=["Module"], path=pathlib.Path("/path/to/venv")
                ),
                "module load Module",
            ),
            # Test case: Apptainer with 1 module and precommand
            (
                testing.config._create_apptainer_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node="a precommand",
                ),
                "a precommand; module load Module",
            ),
            # Test case: Apptainer with 2 modules and no precommand
            (
                testing.config._create_apptainer_environment(
                    modules=["Module"], path=pathlib.Path("/path/to/venv")
                ),
                "module load Module",
            ),
            # Test case: Apptainer with 2 modules and precommand
            (
                testing.config._create_apptainer_environment(
                    modules=["Module1", "Module2"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node="a precommand",
                ),
                "a precommand; module load Module1 Module2",
            ),
            # Test case: Python with no modules and no precommand
            (
                testing.config._create_python_environment(
                    modules=None, path=pathlib.Path("/path/to/venv")
                ),
                None,
            ),
            # Test case: Python with no modules and precommand
            (
                testing.config._create_python_environment(
                    modules=None,
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node="a precommand",
                ),
                "a precommand",
            ),
            # Test case: Python with 1 module and no precommand
            (
                testing.config._create_python_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                ),
                ("module load Module"),
            ),
            # Test case: Python with 1 module and precommand
            (
                testing.config._create_python_environment(
                    modules=["Module"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node="a precommand",
                ),
                ("a precommand; " "module load Module"),
            ),
            # Test case: Python with 2 modules and no precommand
            (
                testing.config._create_python_environment(
                    modules=["Module1", "Module2"],
                    path=pathlib.Path("/path/to/venv"),
                ),
                ("module load Module1 Module2"),
            ),
            # Test case: Python with 2 modules and precommand
            (
                testing.config._create_python_environment(
                    modules=["Module1", "Module2"],
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_login_node="a precommand",
                ),
                ("a precommand; module load Module1 Module2"),
            ),
            # Test case: No execution environment
            (
                environment.Environment(
                    modules=["Module1", "Module2"],
                    pre_run_command_on_login_node="a precommand",
                ),
                "a precommand; module load Module1 Module2",
            ),
        ],
    )
    def test_create_precommand_on_login_node(self, env, expected):
        result = env._create_precommand_on_login_node()

        assert expected == result

    def test_create_postcommand_on_login_node(self):
        expected = "a postcommand"
        env = testing.config._create_python_environment(
            path=pathlib.Path("/path/to/venv"),
            post_run_command_on_login_node="a postcommand",
        )

        result = env.post_run_command_on_login_node

        assert expected == result

    @pytest.mark.parametrize(
        (
            "env",
            "expected",
        ),
        [
            # Test case: Apptainer with no precommand and no postcommand
            (
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
                "srun apptainer run venv python main.py test",
            ),
            # Test case: Apptainer with a precommand and a postcommand
            (
                testing.config._create_apptainer_environment(
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node="a precommand",
                    post_run_command_on_compute_node="a postcommand",
                ),
                (
                    "a precommand && srun apptainer run venv "
                    "python main.py test && "
                    "a postcommand"
                ),
            ),
            # Test case: Python with no precommand and no postcommand
            (
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv")
                ),
                "source /path/to/venv/bin/activate && python main.py test",
            ),
            # Test case: Python with a precommand and a postcommand
            (
                testing.config._create_python_environment(
                    path=pathlib.Path("/path/to/venv"),
                    pre_run_command_on_compute_node="a precommand",
                    post_run_command_on_compute_node="a postcommand",
                ),
                (
                    "a precommand && source /path/to/venv/bin/activate && "
                    "python main.py test && a postcommand"
                ),
            ),
        ],
    )
    def test_create_executable_key_job_description(self, env, expected):
        entry_point_arguments = "python main.py test"
        result = env._create_executable_key_job_description(
            entry_point_arguments
        )
        assert result == expected

    @pytest.mark.parametrize(
        "env",
        [
            # Test case: Apptainer with no variables
            testing.config._create_apptainer_environment(
                modules=None,
                path=pathlib.Path("/path/to/venv"),
                include_mlflow_env_vars=False,
            ),
            # Test case: Python with no variables
            testing.config._create_python_environment(
                modules=None,
                path=pathlib.Path("/path/to/venv"),
                include_mlflow_env_vars=False,
            ),
        ],
    )
    def test_add_env_variables(self, env):
        assert env.variables is None
        expected = {"env_var": "ENV_VAR"}
        env.add_env_vars(expected)
        result = env.variables
        assert expected == result


@pytest.mark.parametrize(
    ("config", "expected", "error_message"),
    [
        # Test case: No execution environment given
        (
            {"Resources": {}},
            None,
            None,
        ),
        # Test case: One execution environment given
        (
            {"Apptainer": {"Path": "image.sif"}},
            executable.Apptainer(path=pathlib.Path("image.sif")),
            None,
        ),
        # Test case: Two execution environments given, should raise
        (
            {"Apptainer": {}, "Python": {}},
            exceptions.ConfigValidationError(),
            "Only one execution environment is allowed, "
            "but in config these have been "
            "found: 'Apptainer', 'Python'.",
        ),
    ],
)
def test_get_execution_environment(
    expect_raise_if_exception, config, expected, error_message
):
    with expect_raise_if_exception(expected) as e:
        result = environment._get_execution_environment(config)
        assert result == expected

    if error_message is not None:
        assert str(e.value) == error_message



import pytest

import mantik.utils.urls as urls


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        (
            "http://test-url.com/path//with/double//slashes",
            "https://test-url.com/path/with/double/slashes",
        ),
        # Test case: also works with 3 slashes
        (
            "http://test-url.com///path//with/double//slashes",
            "https://test-url.com/path/with/double/slashes",
        ),
    ],
)
def test_ensure_https_and_remove_double_slashes_from_path(url, expected):
    result = urls.remove_double_slashes_from_path(url)

    assert result == expected


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        (
            "http://test.domain.de",
            "https://replaced.domain.de",
        ),
        (
            "https://test.domain.de",
            "https://replaced.domain.de",
        ),
        (
            "http://www.test.domain.de",
            "https://www.replaced.domain.de",
        ),
        (
            "https://www.test.domain.de",
            "https://www.replaced.domain.de",
        ),
        (
            "https://www.test.another.domain.de",
            "https://www.replaced.another.domain.de",
        ),
    ],
)
def test_replace_first_subdomain(url, expected):
    result = urls.replace_first_subdomain(url, replace_with="replaced")

    assert result == expected



import pytest

import mlflow_tracking_server.flask.skip as skip


def func_without_skip_flag() -> None:
    pass


@skip.skip_authentication
def func_with_skip_flag() -> None:
    pass


@pytest.mark.parametrize(
    ("func", "expected"),
    [
        (func_without_skip_flag, False),
        (func_with_skip_flag, True),
    ],
)
def test_has_skip_authentication_flag(func, expected):
    result = skip.has_skip_authentication_flag(func)

    assert result == expected



import re
import urllib.parse


def remove_double_slashes_from_path(url: str, ensure_https: bool = True) -> str:
    """Ensure that a URL uses HTTPS and does not contain `//` in its path."""
    u = urllib.parse.urlparse(url)
    path = u.path.replace("//", "/")
    url = f"{u.netloc}{path}".replace("//", "/")
    if ensure_https:
        return f"https://{url}"
    return f"{u.scheme}://{url}"


def replace_first_subdomain(url: str, replace_with: str) -> str:
    regex = re.compile(r"^https?:\/\/(www\.)?(.*?)(\..*)$")
    return regex.sub(rf"https://\1{replace_with}\3", url)



import dataclasses
import pathlib
import typing as t
import uuid

import mantik.utils.mantik_api.client as client


@dataclasses.dataclass
class PostPutDataRepository:
    uri: str
    data_repository_name: t.Optional[str] = None
    access_token: t.Optional[str] = None
    description: t.Optional[str] = None

    @classmethod
    def from_target(
        cls,
        target: t.Union[str, pathlib.Path],
        user: str,
        data_repository_name: t.Optional[str],
        data_repository_description: t.Optional[str],
    ):
        return cls(
            uri=str(target),
            data_repository_name=data_repository_name
            if data_repository_name
            else str(target),
            description=data_repository_description
            if data_repository_description
            else (
                f"This data repository with path: {str(target)} "
                f"was uploaded by {user}"
            ),
        )

    def to_dict(self) -> t.Dict:
        return {
            "dataRepositoryName": self.data_repository_name,
            "uri": self.uri,
            "accessToken": self.access_token,
            "description": self.description,
        }


def add(
    add_data_repository: PostPutDataRepository,
    project_id: uuid.UUID,
    token: str,
):
    data = add_data_repository.to_dict()
    endpoint = f"/mantik-api/projects/{str(project_id)}/data"
    response = client.send_request_to_mantik_api(
        method="POST", data=data, url_endpoint=endpoint, token=token
    )
    client.logger.info(
        f'A new data repository with ID: {response.json()["dataRepositoryId"]} '
        f"and name: {add_data_repository.data_repository_name} at "
        f"{add_data_repository.uri} has been created"
    )
    return response.json()["dataRepositoryId"]


def delete(
    project_id: uuid.UUID,
    data_repository_id: uuid.UUID,
    token: str,
) -> None:
    data = {}
    endpoint = (
        f"/mantik-api/projects/{str(project_id)}/data/{str(data_repository_id)}"
    )
    client.send_request_to_mantik_api(
        method="DELETE", data=data, url_endpoint=endpoint, token=token
    )
    client.logger.info(
        f"Data repository with ID: {data_repository_id} has been deleted"
    )


def get_all(
    project_id: uuid.UUID,
    token: str,
) -> t.List[t.Dict]:
    endpoint = f"/mantik-api/projects/{str(project_id)}/data"
    response = client.send_request_to_mantik_api(
        method="GET", data={}, url_endpoint=endpoint, token=token
    )
    return response.json()["dataRepositories"]



import os
import unittest.mock
import uuid

import click.testing
import pytest

import mantik.cli.main as main
import mantik.cli.models.models

GROUP_NAME = mantik.cli.models.models.GROUP_NAME


SAMPLE_UUID = uuid.uuid4()


@pytest.fixture
def sample_uuid_str() -> str:
    return str(SAMPLE_UUID)


@pytest.fixture
def mock_authentication():
    with unittest.mock.patch(
        "mantik.authentication.auth.get_valid_access_token",
        return_value="1234-token",
    ):
        yield


@pytest.fixture
def mock_project_id_env_var(sample_uuid_str):
    os.environ[mantik.cli.models.models._PROJECT_ID_ENV_VAR] = sample_uuid_str
    yield
    os.unsetenv(mantik.cli.models.models._PROJECT_ID_ENV_VAR)


@pytest.fixture
def cli_test_runner():
    yield click.testing.CliRunner()


def test_get_all_project_models(
    mock_authentication, cli_test_runner, sample_uuid_str
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.get_all"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [GROUP_NAME, "list", "--project-id", sample_uuid_str],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_get_model(mock_authentication, cli_test_runner, sample_uuid_str):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.get_one"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "get-one",
                "--model-id",
                sample_uuid_str,
                "--project-id",
                sample_uuid_str,
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_delete_model(
    mock_authentication,
    mock_project_id_env_var,
    cli_test_runner,
    sample_uuid_str,
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.delete"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [GROUP_NAME, "delete", "--model-id", sample_uuid_str],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_create_model_entry(
    mock_authentication, cli_test_runner, sample_uuid_str
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.add"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "add",
                "--project-id",
                sample_uuid_str,
                "--uri",
                "URI",
                "--location",
                "something://somewhere.com",
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


def test_update_model_entry(
    mock_authentication, cli_test_runner, sample_uuid_str
):
    with unittest.mock.patch(
        "mantik.utils.mantik_api.models.update"
    ) as mocked_function:
        result = cli_test_runner.invoke(
            main.cli,
            [
                GROUP_NAME,
                "update",
                "--project-id",
                sample_uuid_str,
                "--model-id",
                sample_uuid_str,
                "--uri",
                "URI",
                "--location",
                "something://somewhere.com",
            ],
        )
        assert result.exit_code == 0
        mocked_function.assert_called()


@pytest.mark.parametrize(
    "mlflow_params,expected",
    [
        ('{"hello":"world"}', {"hello": "world"}),
        (None, None),
    ],
)
def test_mlflow_parameters_decoder(mlflow_params, expected):
    assert (
        mantik.cli.models.models.mlflow_parameters_decoder(
            None, None, mlflow_params
        )
        == expected
    )


def test_mlflow_parameters_decoder_raises_error_on_wrong_format():
    with pytest.raises(TypeError):
        mantik.cli.models.models.mlflow_parameters_decoder(
            None, None, "i am a broken json string"
        )



import pathlib

import pytest


@pytest.fixture(scope="session")
def mlproject_path() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent
        / "../../../../tests/resources/test-project"
    )



import contextlib
import typing as t

import flask
import pytest
import requests_mock

import mantik.testing as testing
import mantik.utils.env as env
import mantik.utils.mantik_api.client as mantik_api_client
import mlflow_tracking_server.flask.app as flask_app
import tokens.verifier as verifier


@pytest.fixture()
def app(monkeypatch) -> flask.Flask:
    monkeypatch.setattr(
        verifier,
        "TokenVerifier",
        testing.mlflow_server.FakeTokenVerifier,
    )

    _app = flask_app.app
    _app.config.update(
        {
            "TESTING": True,
        }
    )

    yield _app


@pytest.fixture()
def client(app):
    return app.test_client()


@pytest.fixture()
def api_path() -> str:
    return "/api/2.0/mlflow"


@pytest.fixture()
def api_base_url() -> str:
    return "https://api.dev.cloud.mantik.ai/"


@pytest.fixture()
@contextlib.contextmanager
def set_unset_api_base_url(api_base_url):
    env.set_env_vars({mantik_api_client._MANTIK_API_URL_ENV_VAR: api_base_url})
    yield
    env.unset_env_vars([mantik_api_client._MANTIK_API_URL_ENV_VAR])


@pytest.fixture()
def mock_api_permissions(set_unset_api_base_url, api_base_url):
    class Mocker:
        def __init__(self, mocked_urls: t.List[t.Tuple[str, str, t.Dict]]):
            """
            mocked_urls should be a list of tuples that contain:
            method: str e.g. GET, POST ...
            url: str e.g. mantik-api/mlflow/permissions/experiments"
            json: dict e.g. {
                    "allowed": True,
                    "message": "This is a mock test",
                    "experimentIds": None,
                }
            """
            self.mocker = requests_mock.Mocker(real_http=True)
            for method, url, json in mocked_urls:
                self.mocker.register_uri(
                    method,
                    api_base_url + url,
                    json=json,
                )

        def __enter__(self):
            env.set_env_vars(
                {mantik_api_client._MANTIK_API_URL_ENV_VAR: api_base_url}
            )
            return self.mocker.__enter__()

        def __exit__(self, type, value, traceback):
            env.unset_env_vars([mantik_api_client._MANTIK_API_URL_ENV_VAR])
            self.mocker.__exit__(type, value, traceback)

    return Mocker


@pytest.fixture()
def add_experiments_to_mlflow(client, mock_api_permissions):
    """
    Return the experiment ids of the added experiments
    """

    def wrapper(experiment_names: t.List[str]):
        with mock_api_permissions(
            [
                (
                    "POST",
                    "mantik-api/mlflow/permissions/experiments",
                    {
                        "allowed": True,
                        "message": "Allow for anything",
                        "experimentIds": None,
                    },
                )
            ]
        ):
            for name in experiment_names:
                response = client.post(
                    "api/2.0/mlflow/experiments/create",
                    json={"name": name},
                    headers={"Authorization": "test-valid-token"},
                )

                assert (
                    response.status_code == 200
                    or response.json["error_code"] == "RESOURCE_ALREADY_EXISTS"
                )

    return wrapper


@pytest.fixture()
def get_experiment_ids(client, mock_api_permissions):
    def wrapper(experiment_names: t.List[str]):
        with mock_api_permissions(
            [
                (
                    "POST",
                    "mantik-api/mlflow/permissions/experiments",
                    {
                        "allowed": True,
                        "message": "Allow for anything",
                        "experimentIds": None,
                    },
                )
            ]
        ):
            return [
                client.get(
                    "api/2.0/mlflow/experiments/get-by-name",
                    json={"experiment_name": name},
                    headers={"Authorization": "test-valid-token"},
                ).json["experiment"]["experiment_id"]
                for name in experiment_names
            ]

    return wrapper



import datetime
import json
import os
import pathlib

import pytest

import mantik.compute_backend as compute_backend
import mantik.compute_backend.config.core as core
import mantik.unicore as unicore
import mantik.unicore.credentials as unicore_credentials
import mantik.utils as utils

FILE_DIR = pathlib.Path(__file__).parent


@pytest.fixture
def required_config_env_vars():
    return {
        unicore_credentials._USERNAME_ENV_VAR: "test-user",
        unicore_credentials._PASSWORD_ENV_VAR: "test-password",
        core._PROJECT_ENV_VAR: "test-project",
    }


@pytest.fixture(scope="session")
def mlproject_path() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent
        / "../../../../../tests/resources/test-project"
    )


@pytest.fixture(scope="session")
def invalid_config_type() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent
        / "../../../../../tests/resources/broken-project/compute-backend-config.md"  # noqa: E501
    )


@pytest.fixture(scope="session")
def config_with_errors() -> pathlib.Path:
    return (
        pathlib.Path(__file__).parent / "../../../../../tests/resources/"
        "test-project/config-with-errors.yaml"
    )


@pytest.fixture(scope="session")
def compute_backend_config_yaml(mlproject_path) -> pathlib.Path:
    """Return the UNICORE config in YAML format.

    Doesn't contain. `Environment` section

    """
    return pathlib.Path(f"{str(mlproject_path)}/compute-backend-config.yaml")


@pytest.fixture(scope="session")
def compute_backend_config_json(mlproject_path) -> pathlib.Path:
    return pathlib.Path(f"{str(mlproject_path)}/compute-backend-config.json")


@pytest.fixture()
def unset_tracking_token_env_var_before_execution():
    if utils.mlflow.TRACKING_TOKEN_ENV_VAR in os.environ:
        del os.environ[utils.mlflow.TRACKING_TOKEN_ENV_VAR]


@pytest.fixture(scope="function")
def example_config() -> compute_backend.config.core.Config:
    return compute_backend.config.core.Config(
        api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=compute_backend.config.environment.Environment(
            execution=compute_backend.config.executable.Apptainer(
                path=pathlib.Path("mantik-test.sif"),
            ),
            variables={"SRUN_CPUS_PER_TASK": 100},
        ),
        resources=compute_backend.config.resources.Resources(queue="batch"),
        exclude=["*.py", "*.sif"],
    )


@pytest.fixture()
def example_job_property_response() -> dict:
    with open(
        FILE_DIR
        / "../../../resources/unicore-responses/job-property-response.json",
    ) as f:
        return json.load(f)


@pytest.fixture()
def example_job_properties():
    return unicore.properties.Properties(
        status=unicore.properties.Status.SUCCESSFUL,
        logs=[],
        owner="owner",
        site_name="siteName",
        consumed_time=unicore.properties.ConsumedTime(
            total=datetime.timedelta(seconds=1),
            queued=datetime.timedelta(seconds=2),
            stage_in=datetime.timedelta(seconds=3),
            pre_command=datetime.timedelta(seconds=4),
            main=datetime.timedelta(seconds=5),
            post_command=datetime.timedelta(seconds=6),
            stage_out=datetime.timedelta(seconds=7),
        ),
        current_time=datetime.datetime(
            2000, 1, 1, tzinfo=unicore.properties.TZ_OFFSET
        ),
        submission_time=datetime.datetime(
            2000, 1, 1, tzinfo=unicore.properties.TZ_OFFSET
        ),
        termination_time=datetime.datetime(
            2000, 1, 2, tzinfo=unicore.properties.TZ_OFFSET
        ),
        status_message="statusMessage",
        tags=["tag"],
        resource_status="resourceStatus",
        name="name",
        exit_code="0",
        queue="queue",
        submission_preferences={"any": "preferences"},
        resource_status_message="resourceStatusMessage",
        acl=["acl"],
        batch_system_id="batchSystemID",
    )


@pytest.fixture()
def example_config_for_python() -> compute_backend.config.core.Config:
    return compute_backend.config.core.Config(
        api_url="test-url",
        user="user",
        password="password",
        project="test-project",
        environment=compute_backend.config.environment.Environment(
            execution=compute_backend.config.executable.Python(
                path=pathlib.Path("/venv"),
            )
        ),
        resources=compute_backend.config.resources.Resources(queue="batch"),
        exclude=["*.sif"],
    )



import logging
import typing as t
import uuid

import mantik.authentication.auth
import mantik.remote_file_service.abstract_file_service as afs
import mantik.utils.mantik_api.data_repository as data_repository

_PROJECT_ID_ENV_VAR = "MANTIK_PROJECT_ID"

logger = logging.getLogger(__name__)


class DataClientException(Exception):
    pass


class DataClient(afs.AbstractFileService):
    def __init__(
        self,
        project_id: t.Union[str, uuid.UUID],
        file_service: afs.AbstractFileService,
    ):
        self.access_token = mantik.authentication.auth.get_valid_access_token()
        self.project_id = _str_or_uuid_to_uuid(project_id)
        self.file_service = file_service

    @classmethod
    def from_env(
        cls,
        connection_id: t.Optional[uuid.UUID] = None,
        remote_fs_type: t.Optional[t.Type[afs.AbstractFileService]] = None,
    ) -> "DataClient":
        if remote_fs_type is None:
            supported_types = _get_available_remote_file_systems()
            remote_fs_type_from_env = mantik.utils.env.get_required_env_var(
                afs.REMOTE_FS_TYPE_ENV_VAR
            )
            try:
                remote_fs_type = supported_types[remote_fs_type_from_env]
            except KeyError:
                raise DataClientException(
                    "Invalid remote file system type, set REMOTE_FS_TYPE"
                    f"as one of this supported types {list(supported_types.keys())}"  # noqa
                )

        return cls(
            project_id=_str_or_uuid_to_uuid(
                mantik.utils.env.get_required_env_var(_PROJECT_ID_ENV_VAR)
            ),
            file_service=remote_fs_type.from_env(connection_id=connection_id),
        )

    def list_directory(
        self, target: afs.FilePath
    ) -> t.List[t.Union[afs.Directory, afs.File]]:
        return self.file_service.list_directory(target)

    def create_directory(
        self,
        target: afs.FilePath,
        data_repository_name: t.Optional[str] = None,
        data_repository_description: t.Optional[str] = None,
    ) -> afs.Directory:
        directory = self.file_service.create_directory(target)

        self._create_data_reference(
            data_repository.PostPutDataRepository.from_target(
                target=target,
                user=self.user,
                data_repository_name=data_repository_name,
                data_repository_description=data_repository_description,
            )
        )

        return directory

    def remove_directory(
        self,
        target: afs.FilePath,
        data_repository_id: t.Optional[t.Union[str, uuid.UUID]] = None,
        data_repository_name: t.Optional[str] = None,
    ) -> None:
        self._delete_data_reference(
            data_repository_id=data_repository_id,
            data_repository_name=data_repository_name,
        )
        self.file_service.remove_directory(target)

    def copy_directory(
        self,
        source: afs.FilePath,
        target: afs.FilePath,
        data_repository_name: t.Optional[str] = None,
        data_repository_description: t.Optional[str] = None,
    ) -> afs.Directory:
        directory = self.file_service.copy_directory(source, target)

        self._create_data_reference(
            data_repository.PostPutDataRepository.from_target(
                target=target,
                user=self.user,
                data_repository_name=data_repository_name,
                data_repository_description=data_repository_description,
            )
        )

        return directory

    def create_file_if_not_exists(
        self,
        target=afs.FilePath,
        data_repository_name: t.Optional[str] = None,
        data_repository_description: t.Optional[str] = None,
    ) -> afs.File:
        file = self.file_service.create_file_if_not_exists(target)

        self._create_data_reference(
            data_repository.PostPutDataRepository.from_target(
                target=target,
                user=self.user,
                data_repository_name=data_repository_name,
                data_repository_description=data_repository_description,
            )
        )
        return file

    def remove_file(
        self,
        target: afs.FilePath,
        data_repository_id: t.Optional[t.Union[str, uuid.UUID]] = None,
        data_repository_name: t.Optional[str] = None,
    ) -> None:
        self._delete_data_reference(
            data_repository_id=data_repository_id,
            data_repository_name=data_repository_name,
        )
        self.file_service.remove_file(target)

    def copy_file(
        self,
        source: afs.FilePath,
        target: afs.FilePath,
        data_repository_name: t.Optional[str] = None,
        data_repository_description: t.Optional[str] = None,
    ) -> afs.File:
        file = self.file_service.copy_file(source, target)

        self._create_data_reference(
            data_repository.PostPutDataRepository.from_target(
                target=target,
                user=self.user,
                data_repository_name=data_repository_name,
                data_repository_description=data_repository_description,
            )
        )

        return file

    def exists(self, target=afs.FilePath) -> bool:
        return self.file_service.exists(target=target)

    @property
    def user(self) -> str:
        return self.file_service.user

    def change_permissions(
        self, target: afs.FilePath, new_permissions: afs.FileMeta
    ) -> None:
        self.file_service.change_permissions(target, new_permissions)

    def _create_data_reference(
        self,
        add_data_repository: data_repository.PostPutDataRepository,
    ):
        try:
            data_repository_id = data_repository.add(
                add_data_repository=add_data_repository,
                project_id=self.project_id,
                token=self.access_token,
            )
        except Exception as e:
            raise DataClientException(
                "During the creation of a new data repository,"
                f"the following error occurred: {str(e)}\n"
                "The data has been anyway "
                "successfully transferred remotely."
            )
        return data_repository_id

    def _delete_data_reference(
        self,
        data_repository_id: t.Optional[t.Union[str, uuid.UUID]] = None,
        data_repository_name: t.Optional[str] = None,
    ):
        if data_repository_id:
            data_repository_id = _str_or_uuid_to_uuid(data_repository_id)
            data_repository.delete(
                project_id=self.project_id,
                data_repository_id=data_repository_id,
                token=self.access_token,
            )
            return None
        if data_repository_name:
            for data_repo in data_repository.get_all(
                self.project_id, self.access_token
            ):
                if data_repo.get("dataRepositoryName") == data_repository_name:
                    data_repository.delete(
                        project_id=self.project_id,
                        data_repository_id=uuid.UUID(
                            data_repo["dataRepositoryId"]
                        ),
                        token=self.access_token,
                    )
                    return None
            raise DataClientException(
                f"A data repository with name: "
                f"{data_repository_name} was not found.\n"
                "Make sure that the name is correct and "
                "that the repository has not been deleted already.\n"
                "To solve the problem, consider passing 'data_repository_id'."
            )
        raise DataClientException(
            "Either 'data_repository_id' or 'data_repository_name' "
            "must be passed as arguments, "
            "in order to delete the associated data repository in Mantik.\n"
            "If no reference is present use the FileService directly."
        )


def _str_or_uuid_to_uuid(id_: t.Union[str, uuid.UUID]):
    if isinstance(id_, str):
        try:
            id_ = uuid.UUID(id_)
        except ValueError:
            raise DataClientException(
                "Badly formed hexadecimal UUID string for project ID"
            )
    return id_


def _get_available_remote_file_systems() -> t.Dict:
    supported_types = {}
    _attempt_file_service_import(
        supported_types=supported_types,
        name="S3",
        module="s3_file_service",
        cls="S3FileService",
        extras="s3",
    )
    _attempt_file_service_import(
        supported_types=supported_types,
        name="UNICORE",
        module="unicore_file_service",
        cls="UnicoreFileService",
        extras="s3",
    )
    return supported_types


def _attempt_file_service_import(
    supported_types: t.Dict, name: str, module: str, cls: str, extras: str
) -> None:
    """Attempt to import file service from module.

    Parameters
    ----------
    supported_types : dict
        The supported remote file system types.

        If available, the file system will be added.
    name : str
        Name of the remote file system type.

        E.g. ``"S3"`` or ``"UNICORE"``.
    module : str
        Name of the module inside ``mantik.remote_file_system``.

        E.g. ``"s3_file_service"`` or ``"unicore_file_service"``.
    cls : str
        Name of the class inside the respective module.
    extras : str
        Name of the extras inside ``pyproject.toml`` required for
        the remote file system.

    Notes
    -----
    This is required to avoid import errors when extras are not installed.

    Extras for the respective file services are defined in ``pyproject.toml``.

    """
    try:
        sub_module = getattr(mantik.remote_file_service, module)
    except ModuleNotFoundError:
        logger.warning(
            (
                "Unable to import %s remote file service, "
                'install required requirements via `pip install "mantik[%s]"`'
            ),
            name,
            extras,
            exc_info=True,
        )
    else:
        supported_types[name] = getattr(sub_module, cls)



import dataclasses
import typing as t

import mantik.compute_backend.config._base as _base
import mantik.compute_backend.config._utils as _utils


@dataclasses.dataclass
class Resources(_base.ConfigObject):
    """The computing resources that will be requested ."""

    queue: str
    runtime: t.Optional[str] = None
    nodes: t.Optional[int] = None
    total_cpus: t.Optional[int] = None
    cpus_per_node: t.Optional[int] = None
    gpus_per_node: t.Optional[int] = None
    memory_per_node: t.Optional[str] = None
    reservation: t.Optional[str] = None
    node_constraints: t.Optional[str] = None
    qos: t.Optional[str] = None

    @classmethod
    def _from_dict(cls, config: t.Dict) -> "Resources":
        queue = _utils.get_required_config_value(
            name="Queue",
            value_type=str,
            config=config,
        )
        runtime = _utils.get_optional_config_value(
            name="Runtime",
            value_type=str,
            config=config,
        )
        nodes = _utils.get_optional_config_value(
            name="Nodes",
            value_type=int,
            config=config,
        )
        total_cpus = _utils.get_optional_config_value(
            name="TotalCPUs",
            value_type=int,
            config=config,
        )
        cpus_per_node = _utils.get_optional_config_value(
            name="CPUsPerNode",
            value_type=int,
            config=config,
        )
        gpus_per_node = _utils.get_optional_config_value(
            name="GPUsPerNode",
            value_type=int,
            config=config,
        )
        memory_per_node = _utils.get_optional_config_value(
            name="MemoryPerNode",
            value_type=str,
            config=config,
        )
        reservation = _utils.get_optional_config_value(
            name="Reservation",
            value_type=str,
            config=config,
        )
        node_constraints = _utils.get_optional_config_value(
            name="NodeConstraints",
            value_type=str,
            config=config,
        )
        qos = _utils.get_optional_config_value(
            name="QoS",
            value_type=str,
            config=config,
        )
        return cls(
            queue=queue,
            runtime=runtime,
            nodes=nodes,
            total_cpus=total_cpus,
            cpus_per_node=cpus_per_node,
            gpus_per_node=gpus_per_node,
            memory_per_node=memory_per_node,
            reservation=reservation,
            node_constraints=node_constraints,
            qos=qos,
        )

    def _to_dict(self) -> t.Dict:
        return {
            "Runtime": self.runtime,
            "Queue": self.queue,
            "Nodes": self.nodes,
            "TotalCPUs": self.total_cpus,
            "CPUsPerNode": self.cpus_per_node,
            "GPUS": self.gpus_per_node,
            "MemoryPerNode": self.memory_per_node,
            "Reservation": self.reservation,
            "NodeConstraints": self.node_constraints,
            "QoS": self.qos,
        }
